#!/usr/bin/env python

import asyncio
import functools
import logging
import threading
import time
from collections import deque
from dataclasses import asdict, is_dataclass
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

from hummingbot import get_logging_conf
from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:  # pragma: no cover
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401
    from hummingbot.core.event.event_listener import EventListener  # noqa: F401

from commlib.node import Node, NodeState
from commlib.transports.mqtt import ConnectionParameters as MQTTConnectionParameters

from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, DeductedFromReturnsTradeFee
from hummingbot.core.event import events
from hummingbot.core.event.event_forwarder import SourceInfoEventForwarder
from hummingbot.core.pubsub import PubSub
from hummingbot.core.utils.async_utils import call_sync, safe_ensure_future
from hummingbot.notifier.notifier_base import NotifierBase
from hummingbot.remote_iface.messages import (
    MQTT_STATUS_CODE,
    BalanceLimitCommandMessage,
    BalancePaperCommandMessage,
    CommandShortcutMessage,
    ConfigCommandMessage,
    ExternalEventMessage,
    HistoryCommandMessage,
    ImportCommandMessage,
    InternalEventMessage,
    LogMessage,
    NotifyMessage,
    StartCommandMessage,
    StatusCommandMessage,
    StatusUpdateMessage,
    StopCommandMessage,
)

mqtts_logger: HummingbotLogger = None


class CommandTopicSpecs:
    START: str = '/start'
    STOP: str = '/stop'
    CONFIG: str = '/config'
    IMPORT: str = '/import'
    STATUS: str = '/status'
    HISTORY: str = '/history'
    BALANCE_LIMIT: str = '/balance/limit'
    BALANCE_PAPER: str = '/balance/paper'
    COMMAND_SHORTCUT: str = '/command_shortcuts'


class TopicSpecs:
    PREFIX: str = '{namespace}/{instance_id}'
    COMMANDS: CommandTopicSpecs = CommandTopicSpecs()
    LOGS: str = '/log'
    INTERNAL_EVENTS: str = '/events'
    NOTIFICATIONS: str = '/notify'
    STATUS_UPDATES: str = '/status_updates'
    HEARTBEATS: str = '/hb'
    EXTERNAL_EVENTS: str = '/external/event/*'


class MQTTCommands:
    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            raise EnvironmentError(
                "MQTTCommands can only be initialized from the main thread."
            )
        self._hb_app = hb_app
        self._node = node
        self.logger = self._hb_app.logger
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._start_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.START}'
        self._stop_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.STOP}'
        self._config_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.CONFIG}'
        self._import_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.IMPORT}'
        self._status_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.STATUS}'
        self._history_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.HISTORY}'
        self._balance_limit_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.BALANCE_LIMIT}'
        self._balance_paper_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.BALANCE_PAPER}'
        self._shortcuts_uri = f'{topic_prefix}{TopicSpecs.COMMANDS.COMMAND_SHORTCUT}'

        self._init_commands()

    def _init_commands(self):
        self._node.create_rpc(
            rpc_name=self._start_uri,
            msg_type=StartCommandMessage,
            on_request=self._on_cmd_start
        )
        self._node.create_rpc(
            rpc_name=self._stop_uri,
            msg_type=StopCommandMessage,
            on_request=self._on_cmd_stop
        )
        self._node.create_rpc(
            rpc_name=self._config_uri,
            msg_type=ConfigCommandMessage,
            on_request=self._on_cmd_config
        )
        self._node.create_rpc(
            rpc_name=self._import_uri,
            msg_type=ImportCommandMessage,
            on_request=self._on_cmd_import
        )
        self._node.create_rpc(
            rpc_name=self._status_uri,
            msg_type=StatusCommandMessage,
            on_request=self._on_cmd_status
        )
        self._node.create_rpc(
            rpc_name=self._history_uri,
            msg_type=HistoryCommandMessage,
            on_request=self._on_cmd_history
        )
        self._node.create_rpc(
            rpc_name=self._balance_limit_uri,
            msg_type=BalanceLimitCommandMessage,
            on_request=self._on_cmd_balance_limit
        )
        self._node.create_rpc(
            rpc_name=self._balance_paper_uri,
            msg_type=BalancePaperCommandMessage,
            on_request=self._on_cmd_balance_paper
        )
        self._node.create_rpc(
            rpc_name=self._shortcuts_uri,
            msg_type=CommandShortcutMessage,
            on_request=self._on_cmd_command_shortcut
        )

    def _on_cmd_start(self, msg: StartCommandMessage.Request):
        response = StartCommandMessage.Response()
        timeout = 30
        try:
            if self._hb_app.strategy_name is None and msg.script is None:
                raise Exception('Strategy check: Please import or create a strategy.')
            if self._hb_app.strategy is not None:
                raise Exception('The bot is already running - please run "stop" first')
            if msg.async_backend:
                self._hb_app.start(
                    log_level=msg.log_level,
                    script=msg.script,
                    conf=msg.conf,
                    is_quickstart=msg.is_quickstart
                )
            else:
                res = call_sync(
                    self._hb_app.start_check(
                        log_level=msg.log_level,
                        script=msg.script,
                        conf=msg.conf,
                        is_quickstart=msg.is_quickstart
                    ),
                    loop=self._ev_loop,
                    timeout=timeout
                )
                response.msg = res if res is not None else ''
        except asyncio.exceptions.TimeoutError:
            response.msg = f'Hummingbot start command timed out after {timeout} seconds'
            response.status = MQTT_STATUS_CODE.ERROR
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_stop(self, msg: StopCommandMessage.Request):
        response = StopCommandMessage.Response()
        timeout = 30
        try:
            if msg.async_backend:
                self._hb_app.stop(
                    skip_order_cancellation=msg.skip_order_cancellation
                )
            else:
                res = call_sync(
                    self._hb_app.stop_loop(),
                    loop=self._ev_loop,
                    timeout=timeout
                )
                response.msg = res if res is not None else ''
        except asyncio.exceptions.TimeoutError:
            response.msg = f'Hummingbot start command timed out after {timeout} seconds'
            response.status = MQTT_STATUS_CODE.ERROR
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_config(self, msg: ConfigCommandMessage.Request):
        response = ConfigCommandMessage.Response()
        try:
            if len(msg.params) == 0:
                self._hb_app.config()
            else:
                invalid_params = []
                for param in msg.params:
                    if param[0] in self._hb_app.configurable_keys():
                        self._ev_loop.call_soon_threadsafe(
                            self._hb_app.config,
                            param[0],
                            param[1]
                        )
                        response.changes.append((param[0], param[1]))
                    else:
                        invalid_params.append(param[0])
                if len(invalid_params):
                    raise ValueError(f'Invalid param key(s): {invalid_params}')
            strategy_config = {}
            client_config = {}
            if isinstance(self._hb_app.client_config_map, dict):  # pragma: no cover
                client_config = self._hb_app.client_config_map
                for key, value in client_config.items():
                    if isinstance(value, ConfigVar):
                        client_config[key] = value.value
                    else:
                        client_config[key] = value
            elif isinstance(self._hb_app.client_config_map,
                            ClientConfigAdapter):
                client_config = self._hb_app.client_config_map.dict()
            if isinstance(self._hb_app._strategy_config_map, dict):  # pragma: no cover
                for key, value in self._hb_app._strategy_config_map.items():
                    if isinstance(value, ConfigVar):
                        strategy_config[key] = value.value
                    else:
                        strategy_config[key] = value
            elif isinstance(self._hb_app._strategy_config_map,
                            ClientConfigAdapter):
                strategy_config = self._hb_app._strategy_config_map.dict()
            response.config = {
                "client": client_config,
                "strategy": strategy_config
            }
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
            self._hb_app.logger().error(e)
        return response

    def _on_cmd_import(self, msg: ImportCommandMessage.Request):
        response = ImportCommandMessage.Response()
        timeout = 30  # seconds
        strategy_name = msg.strategy
        if strategy_name in (None, ''):
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = 'Empty strategy_name given!'
            return response
        strategy_file_name = f'{strategy_name}.yml'
        try:
            res = call_sync(
                self._hb_app.import_config_file(strategy_file_name),
                loop=self._ev_loop,
                timeout=timeout
            )
            response.msg = res if res is not None else ''
        except asyncio.exceptions.TimeoutError:
            response.msg = f'Hummingbot import command timed out after {timeout} seconds'
            response.status = MQTT_STATUS_CODE.ERROR
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_status(self, msg: StatusCommandMessage.Request):
        response = StatusCommandMessage.Response()
        timeout = 30  # seconds
        try:
            if self._hb_app.strategy is None:
                response.status = MQTT_STATUS_CODE.ERROR
                response.msg = 'No strategy is currently running!'
                return response
            if msg.async_backend:
                self._ev_loop.call_soon_threadsafe(
                    self._hb_app.status
                )
            else:
                res = call_sync(
                    self._hb_app.strategy_status(),
                    loop=self._ev_loop,
                    timeout=timeout
                )
                response.msg = res if res is not None else ''
        except asyncio.exceptions.TimeoutError:
            response.msg = f'Hummingbot status command timed out after {timeout} seconds'
            response.status = MQTT_STATUS_CODE.ERROR
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_history(self, msg: HistoryCommandMessage.Request):
        response = HistoryCommandMessage.Response()
        try:
            if msg.async_backend:
                self._hb_app.history(msg.days, msg.verbose, msg.precision)
            else:
                trades = self._hb_app.get_history_trades_json(msg.days)
                if trades:
                    response.trades = trades
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_balance_limit(self, msg: BalanceLimitCommandMessage.Request):
        response = BalanceLimitCommandMessage.Response()
        try:
            data = self._hb_app.balance(
                'limit',
                [msg.exchange, msg.asset, msg.amount]
            )
            response.data = data
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_balance_paper(self, msg: BalancePaperCommandMessage.Request):
        response = BalancePaperCommandMessage.Response()
        try:
            data = self._hb_app.balance(
                'paper',
                [msg.asset, msg.amount]
            )
            response.data = data
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response

    def _on_cmd_command_shortcut(self, msg: CommandShortcutMessage.Request):
        response = CommandShortcutMessage.Response()
        try:
            for param in msg.params:
                response.success.append(self._hb_app._handle_shortcut(param))
        except Exception as e:
            response.status = MQTT_STATUS_CODE.ERROR
            response.msg = str(e)
        return response


class MQTTMarketEventForwarder:
    @classmethod
    def logger(cls) -> HummingbotLogger:
        global mqtts_logger
        if mqtts_logger is None:  # pragma: no cover
            mqtts_logger = HummingbotLogger(__name__)
        return mqtts_logger

    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            raise EnvironmentError(
                "MQTTMarketEventForwarder can only be initialized from the main thread."
            )
        self._hb_app = hb_app
        self._node = node
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop
        self._markets: List[ConnectorBase] = list(self._hb_app.markets.values())

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._topic = f'{topic_prefix}{TopicSpecs.INTERNAL_EVENTS}'

        self._mqtt_fowarder: SourceInfoEventForwarder = \
            SourceInfoEventForwarder(self._send_mqtt_event)
        self._market_event_pairs: List[Tuple[int, EventListener]] = [
            (events.MarketEvent.BuyOrderCreated, self._mqtt_fowarder),
            (events.MarketEvent.BuyOrderCompleted, self._mqtt_fowarder),
            (events.MarketEvent.SellOrderCreated, self._mqtt_fowarder),
            (events.MarketEvent.SellOrderCompleted, self._mqtt_fowarder),
            (events.MarketEvent.OrderFilled, self._mqtt_fowarder),
            (events.MarketEvent.OrderFailure, self._mqtt_fowarder),
            (events.MarketEvent.OrderCancelled, self._mqtt_fowarder),
            (events.MarketEvent.OrderExpired, self._mqtt_fowarder),
            (events.MarketEvent.FundingPaymentCompleted, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionLiquidityAdded, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionLiquidityRemoved, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionUpdate, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionUpdateFailure, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionFeeCollected, self._mqtt_fowarder),
            (events.MarketEvent.RangePositionClosed, self._mqtt_fowarder),
        ]

        self.event_fw_pub = self._node.create_publisher(
            topic=self._topic, msg_type=InternalEventMessage
        )
        self._start_event_listeners()

    def _send_mqtt_event(self, event_tag: int, pubsub: PubSub, event):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(
                self._send_mqtt_event,
                event_tag,
                pubsub,
                event
            )
            return
        try:
            event_types = {
                events.MarketEvent.BuyOrderCreated.value: "BuyOrderCreated",
                events.MarketEvent.BuyOrderCompleted.value: "BuyOrderCompleted",
                events.MarketEvent.SellOrderCreated.value: "SellOrderCreated",
                events.MarketEvent.SellOrderCompleted.value: "SellOrderCompleted",
                events.MarketEvent.OrderFilled.value: "OrderFilled",
                events.MarketEvent.OrderCancelled.value: "OrderCancelled",
                events.MarketEvent.OrderExpired.value: "OrderExpired",
                events.MarketEvent.OrderFailure.value: "OrderFailure",
                events.MarketEvent.FundingPaymentCompleted.value: "FundingPaymentCompleted",
                events.MarketEvent.RangePositionLiquidityAdded.value: "RangePositionLiquidityAdded",
                events.MarketEvent.RangePositionLiquidityRemoved.value: "RangePositionLiquidityRemoved",
                events.MarketEvent.RangePositionUpdate.value: "RangePositionUpdate",
                events.MarketEvent.RangePositionUpdateFailure.value: "RangePositionUpdateFailure",
                events.MarketEvent.RangePositionFeeCollected.value: "RangePositionFeeCollected",
                events.MarketEvent.RangePositionClosed.value: "RangePositionClosed",
            }
            event_type = event_types[event_tag]
        except KeyError:
            event_type = "Unknown"

        if is_dataclass(event):
            event_data = asdict(event)
        elif isinstance(event, tuple) and hasattr(event, '_fields'):
            event_data = event._asdict()
        else:
            try:
                event_data = dict(event)
            except (TypeError, ValueError):
                event_data = {}

        try:
            timestamp = event_data.pop('timestamp')
        except KeyError:
            timestamp = datetime.now().timestamp()

        event_data = self._make_event_payload(event_data)

        self.event_fw_pub.publish(
            InternalEventMessage(
                timestamp=int(timestamp),
                type=event_type,
                data=event_data
            )
        )

    def _make_event_payload(self, event_data):
        if 'type' in event_data:
            event_data['type'] = str(event_data['type'])
        if 'order_type' in event_data:
            event_data['order_type'] = str(event_data['order_type'])
        if 'trade_type' in event_data:
            event_data['trade_type'] = str(event_data['trade_type'])

        for key, val in event_data.items():
            if isinstance(val, dict):
                self._make_event_payload(val)
            elif isinstance(val, Decimal):
                event_data[key] = float(val)
            elif isinstance(val, DeductedFromReturnsTradeFee):
                event_data[key] = val.to_json()
                self._make_event_payload(event_data[key])
            elif isinstance(val, AddedToCostTradeFee):
                event_data[key] = val.to_json()
                self._make_event_payload(event_data[key])
        return event_data

    def _start_event_listeners(self):
        for market in self._markets:
            for event_pair in self._market_event_pairs:
                market.add_listener(event_pair[0], event_pair[1])
                self.logger().debug(
                    f'Created MQTT bridge for event: {event_pair[0]}, {event_pair[1]}'
                )

    def _stop_event_listeners(self):
        for market in self._markets:
            for event_pair in self._market_event_pairs:
                market.remove_listener(event_pair[0], event_pair[1])


class MQTTNotifier(NotifierBase):
    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node) -> None:
        super().__init__()
        self._node = node
        self._hb_app = hb_app
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._topic = f'{topic_prefix}{TopicSpecs.NOTIFICATIONS}'
        self.notify_pub = self._node.create_publisher(
            topic=self._topic,
            msg_type=NotifyMessage
        )

    def add_msg_to_queue(self, msg: str):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(self.add_msg_to_queue, msg)
            return
        self.notify_pub.publish(NotifyMessage(msg=msg))

    def start(self) -> None:
        return None

    def stop(self) -> None:
        return None


class MQTTStatusUpdates:
    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node) -> None:
        self._node = node
        self._hb_app = hb_app
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._topic = f'{topic_prefix}{TopicSpecs.STATUS_UPDATES}'
        self.status_updates_pub = self._node.create_publisher(
            topic=self._topic,
            msg_type=StatusUpdateMessage
        )

        if self._node.state == NodeState.RUNNING:
            self.status_updates_pub.run()

    def add_msg_to_queue(self, msg: str, msg_type: str = 'hbapp'):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(self.add_msg_to_queue, msg, msg_type)
            return

        self.status_updates_pub.publish(
            StatusUpdateMessage(
                msg=msg,
                type=msg_type,
                timestamp=int(time.time() * 1e3)
            )
        )

    def stop(self):
        self.status_updates_pub.stop()


class MQTTGateway(Node):
    NODE_NAME: str = 'hbot.$instance_id'
    _instance: Optional["MQTTGateway"] = None
    _INTERVAL_HEALTH_CHECK = 1.0
    _INTERVAL_RESTART_SHORT = 5.0
    _INTERVAL_RESTART_LONG = 10.0

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global mqtts_logger
        if mqtts_logger is None:  # pragma: no cover
            mqtts_logger = logging.getLogger(__name__)
        return mqtts_logger

    @classmethod
    def main(cls) -> "MQTTGateway":
        return cls._instance

    def __init__(self,
                 hb_app: "HummingbotApplication",
                 *args, **kwargs
                 ):
        self._health = False
        self._initial_connection_succeeded = False
        self._restarting = False
        self._stop_event_async = asyncio.Event()
        self._notifier: MQTTNotifier = None
        self._status_updates: MQTTStatusUpdates = None
        self._market_events: MQTTMarketEventForwarder = None
        self._commands: MQTTCommands = None
        self._logh: MQTTLogHandler = None
        self._external_events: MQTTExternalEvents = None
        self._hb_app: "HummingbotApplication" = hb_app
        self._ev_loop = self._hb_app.ev_loop
        self._params = self._create_mqtt_params_from_conf()
        self.namespace = self._hb_app.client_config_map.mqtt_bridge.mqtt_namespace
        if self.namespace[-1] in ('/', '.'):
            self.namespace = self.namespace[:-1]

        self._topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self.namespace,
            instance_id=self._hb_app.instance_id
        )
        _hb_topic = f'{self._topic_prefix}{TopicSpecs.HEARTBEATS}'

        super().__init__(
            node_name=self.NODE_NAME.replace('$instance_id', hb_app.instance_id),
            connection_params=self._params,
            heartbeats=True,
            heartbeat_uri=_hb_topic,
            *args,
            **kwargs
        )
        MQTTGateway._instance = self

    @property
    def health(self):
        return self._health

    def _safe_get_log_handlers(self, max_tries=3):  # pragma: no cover
        current_try = 0
        while current_try < max_tries:
            try:
                return list([logging.getLogger(name) for name in logging.root.manager.loggerDict])
            except RuntimeError:
                current_try += 1

        log_keys = logging.root.manager.loggerDict.keys()
        return list([logging.getLogger(name) for name in log_keys])

    def _remove_log_handlers(self):  # pragma: no cover
        loggers = self._safe_get_log_handlers()
        log_conf = get_logging_conf()

        if 'loggers' not in log_conf:
            return

        logs = [key for key, val in log_conf.get('loggers').items()]
        for logger in loggers:
            if 'hummingbot' in logger.name:
                for log in logs:
                    if log in logger.name:
                        self.remove_log_handler(logger)

        self._logh = None

    def _init_logger(self):
        self._logh = MQTTLogHandler(self._hb_app, self)
        self.patch_loggers()

    def patch_loggers(self):  # pragma: no cover
        loggers = self._safe_get_log_handlers()

        log_conf = get_logging_conf()
        if 'root' in log_conf:
            if log_conf.get('root').get('mqtt'):
                self.remove_log_handler(self._get_root_logger())
                self.add_log_handler(self._get_root_logger())

        if 'loggers' not in log_conf:
            return

        log_conf_names = [key for key, val in log_conf.get('loggers').items()]
        loggers_filtered = [logger for logger in loggers if
                            logger.name in log_conf_names]
        loggers_filtered = [logger for logger in loggers_filtered if
                            log_conf.get('loggers').get(logger.name).get('mqtt', False)]

        for logger in loggers_filtered:
            self.remove_log_handler(logger)
            self.add_log_handler(logger)

    def _get_root_logger(self):
        return logging.getLogger()

    def remove_log_handler(self, logger: HummingbotLogger):
        logger.removeHandler(self._logh)

    def add_log_handler(self, logger: HummingbotLogger):
        logger.addHandler(self._logh)

    def _init_notifier(self):
        if self._hb_app.client_config_map.mqtt_bridge.mqtt_notifier:
            self._notifier = MQTTNotifier(self._hb_app, self)
            self._hb_app.notifiers.append(self._notifier)

    def _remove_notifier(self):
        self._hb_app.notifiers.remove(self._notifier) if self._notifier \
            in self._hb_app.notifiers else None

    def _init_status_updates(self):
        self._status_updates = MQTTStatusUpdates(self._hb_app, self)

    def _remove_status_updates(self):
        if self._status_updates is not None:
            self._status_updates.stop()
            self._status_updates = None

    def broadcast_status_update(self, *args, **kwargs):
        if self._status_updates is not None:
            self._status_updates.add_msg_to_queue(*args, **kwargs)

    def _init_commands(self):
        if self._hb_app.client_config_map.mqtt_bridge.mqtt_commands:
            self._commands = MQTTCommands(self._hb_app, self)

    def start_market_events_fw(self):
        # Must be called after loading the strategy.
        # HummingbotApplication._initialize_markets() must be be called before
        if self._hb_app.client_config_map.mqtt_bridge.mqtt_events:
            self._market_events = MQTTMarketEventForwarder(self._hb_app, self)
            if self.state == NodeState.RUNNING:
                self._market_events.event_fw_pub.run()

    def _remove_market_event_listeners(self):
        if self._market_events is not None:
            self._market_events._stop_event_listeners()

    def _init_external_events(self):
        if self._hb_app.client_config_map.mqtt_bridge.mqtt_external_events:
            self._external_events = MQTTExternalEvents(self._hb_app, self)

    def add_external_event_listener(self,
                                    event_name: str,
                                    callback: Callable[[ExternalEventMessage, str], None]):
        if event_name == '*':
            self._external_events.add_global_listener(callback)
        else:
            self._external_events.add_listener(event_name, callback)

    def remove_external_event_listener(self,
                                       event_name: str,
                                       callback: Callable[[ExternalEventMessage, str], None]):
        if event_name == '*':
            self._external_events.remove_global_listener(callback)
        else:
            self._external_events.remove_listener(event_name, callback)

    def _create_mqtt_params_from_conf(self):
        host = self._hb_app.client_config_map.mqtt_bridge.mqtt_host
        port = self._hb_app.client_config_map.mqtt_bridge.mqtt_port
        username = self._hb_app.client_config_map.mqtt_bridge.mqtt_username
        password = self._hb_app.client_config_map.mqtt_bridge.mqtt_password
        ssl = self._hb_app.client_config_map.mqtt_bridge.mqtt_ssl
        conn_params = MQTTConnectionParameters(
            host=host,
            port=int(port),
            username=username,
            password=password,
            ssl=ssl
        )
        return conn_params

    def _check_connections(self) -> bool:
        if self._restarting:
            return False
        for c in self._publishers:
            if not c._transport.is_connected:
                return False
        for c in self._rpc_services:
            if not c._transport.is_connected:
                return False
        # Will use if subscribtions are integrated
        for c in self._subscribers:
            if not c._transport.is_connected:
                return False
        # Will use if rpc clients are integrated
        # for c in self._rpc_clients:
        #     if not c._transport.is_connected:
        #         return False
        return True

    def _start_health_monitoring_loop(self):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(self._start_health_monitoring_loop)
            return
        self._stop_event_async.clear()
        safe_ensure_future(self._monitor_health_loop(),
                           loop=self._ev_loop)

    async def _monitor_health_loop(self):
        while not self._stop_event_async.is_set():
            # Maybe we can include more checks here to determine the health!
            self._health = await self._ev_loop.run_in_executor(
                None, self._check_connections)
            if self.health:
                if not self._initial_connection_succeeded:
                    self._initial_connection_succeeded = True
                    self._hb_app.logger().debug('Monitoring MQTT Gateway health for disconnections.')

                await asyncio.sleep(self._INTERVAL_HEALTH_CHECK)
            elif self._initial_connection_succeeded and not self._stop_event_async.is_set():
                await self._restart_gateway()

    async def _restart_gateway(self):
        self._hb_app.logger().warning('MQTT Gateway is disconnected, attempting to reconnect.')

        try:
            self._restarting = True
            self.stop(False)
            await asyncio.sleep(self._INTERVAL_RESTART_SHORT)

            self._publishers = []
            self._subscribers = []
            self._rpc_services = []
            # self._rpc_clients = []

            self.start(False)
            if self._hb_app.strategy is not None:
                self.start_market_events_fw()

            await asyncio.sleep(self._INTERVAL_RESTART_SHORT)

            self._restarting = False

            self._health = await self._ev_loop.run_in_executor(
                None, self._check_connections)

            if self._health:
                self._hb_app.logger().warning('MQTT Gateway successfully reconnected.')

        except Exception as e:
            self._hb_app.logger().error(f'MQTT Gateway failed to reconnect: {e}. Sleeping 10 seconds before retry.')

        await asyncio.sleep(self._INTERVAL_RESTART_LONG)

    def _stop_health_monitoring_loop(self):
        self._stop_event_async.set()

    def start(self, with_health: bool = True) -> None:
        self._init_logger()
        self._init_notifier()
        self._init_status_updates()
        self._init_commands()
        self._init_external_events()

        if with_health:
            self._start_health_monitoring_loop()

        self.run()
        self.broadcast_status_update("online", msg_type="availability")

    def stop(self, with_health: bool = True):
        self.broadcast_status_update("offline", msg_type="availability")
        super().stop()
        if self._hb_thread:
            self._hb_thread.stop()
        self._remove_status_updates()
        self._remove_notifier()
        self._remove_log_handlers()
        self._remove_market_event_listeners()

        if with_health:
            self._stop_health_monitoring_loop()

    def __del__(self):
        self.stop()


class MQTTLogHandler(logging.Handler):
    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            raise EnvironmentError(
                "MQTTLogHandler can only be initialized from the main thread."
            )
        self._hb_app = hb_app
        self._node = node
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._topic = f'{topic_prefix}{TopicSpecs.LOGS}'

        super().__init__()
        self.name = self.__class__.__name__
        self.log_pub = self._node.create_publisher(topic=self._topic,
                                                   msg_type=LogMessage)

    def emit(self, record: logging.LogRecord):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(self.emit, record)
            return
        msg_str = self.format(record)
        msg = LogMessage(
            timestamp=time.time(),
            msg=msg_str,
            level_no=record.levelno,
            level_name=record.levelname,
            logger_name=record.name

        )
        self.log_pub.publish(msg)


class MQTTExternalEvents:
    def __init__(self,
                 hb_app: "HummingbotApplication",
                 node: Node
                 ):
        self._node: Node = node
        self._hb_app: 'HummingbotApplication' = hb_app
        self._ev_loop: asyncio.AbstractEventLoop = self._hb_app.ev_loop

        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._hb_app.instance_id
        )
        self._topic = f'{topic_prefix}{TopicSpecs.EXTERNAL_EVENTS}'

        self._node.create_psubscriber(
            topic=self._topic,
            msg_type=ExternalEventMessage,
            on_message=self._on_event_arrived
        )
        self._listeners: Dict[
            str,
            List[Callable[ExternalEventMessage, str], None]
        ] = {'*': []}

    def _event_uri_to_name(self, topic: str) -> str:
        return topic.split('event/')[1].replace('/', '.')

    def _on_event_arrived(self,
                          msg: ExternalEventMessage,
                          topic: str
                          ) -> None:
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            self._ev_loop.call_soon_threadsafe(self._on_event_arrived,
                                               msg, topic)
            return
        event_name = self._event_uri_to_name(topic)
        self._hb_app.logger().debug(
            f'Received external event {event_name} -> {msg} - '
            'Broadcasting to listeners...'
        )
        if event_name in self._listeners:
            for fenc in self._listeners[event_name]:
                fenc(msg, event_name)
        for fenc in self._listeners['*']:
            fenc(msg, event_name)

    def add_listener(self,
                     event_name: str,
                     callback: Callable[[ExternalEventMessage, str], None]
                     ) -> None:
        # TODO validate event_name with regex
        if event_name in self._listeners:
            self._listeners.get(event_name).append(callback)
        else:
            self._listeners[event_name] = [callback]

    def remove_listener(self,
                        event_name: str,
                        callback: Callable[[ExternalEventMessage, str], None]
                        ):
        # TODO validate event_name with regex
        if event_name in self._listeners:
            self._listeners.get(event_name).remove(callback)

    def add_global_listener(self,
                            callback: Callable[[ExternalEventMessage, str], None]
                            ):
        if '*' in self._listeners:
            self._listeners.get('*').append(callback)
        else:
            self._listeners['*'] = [callback]

    def remove_global_listener(self,
                               callback: Callable[[ExternalEventMessage, str], None]
                               ):
        if '*' in self._listeners:
            self._listeners.get('*').remove(callback)


class ETopicListener:
    def __init__(self,
                 topic: str,
                 on_message: Callable[[Dict[str, Any], str], None],
                 use_bot_prefix: Optional[bool] = True
                 ):
        self._node = MQTTGateway.main()
        if self._node is None:
            raise Exception('MQTT Gateway not yet initialized')
        topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._node._hb_app.instance_id
        )
        if use_bot_prefix:
            self._topic = f'{topic_prefix}/{topic}'
        else:
            self._topic = topic
        self._on_message = on_message
        self._sub = self._node.create_psubscriber(topic=self._topic,
                                                  on_message=self._on_message)
        if self._node.state == NodeState.RUNNING:
            self._sub.run()

    def stop(self):
        self._sub.stop()


class EEventQueueFactory:
    @classmethod
    def create(cls,
               event_name: str,
               queue_size: Optional[int] = 1000
               ) -> deque:
        gw = MQTTGateway.main()
        queue = deque(maxlen=queue_size)
        if gw is None:
            raise Exception('MQTTGateway is offline!')
        _on_event_clb = functools.partial(cls._on_event, queue)
        gw.add_external_event_listener(event_name, _on_event_clb)
        return queue

    @classmethod
    def _on_event(cls, queue: deque, msg: Dict[str, Any], name):
        queue.append((name, msg))


class EEventListenerFactory:
    @classmethod
    def create(cls,
               event_name: str,
               callback: Callable[[Dict[str, Any], str], None],
               ) -> None:
        gw = MQTTGateway.main()
        if gw is None:
            raise Exception('MQTTGateway is offline!')
        gw.add_external_event_listener(event_name, callback)

    @classmethod
    def remove(cls,
               event_name: str,
               callback: Callable[[Dict[str, Any], str], None],
               ) -> None:
        gw = MQTTGateway.main()
        if gw is None:
            raise Exception('MQTTGateway is offline!')
        gw.remove_external_event_listener(event_name, callback)


class ETopicListenerFactory:
    @classmethod
    def create(cls,
               topic: str,
               callback: Callable[[Dict[str, Any], str], None],
               use_bot_prefix: Optional[bool] = True
               ) -> ETopicListener:
        listener = ETopicListener(
            topic=topic,
            on_message=callback,
            use_bot_prefix=use_bot_prefix
        )
        return listener

    @classmethod
    def remove(cls, listener):
        listener.stop()
        del listener


class ETopicQueueFactory:
    @classmethod
    def create(cls,
               topic: str,
               queue_size: Optional[int] = 1000,
               use_bot_prefix: Optional[bool] = True
               ) -> deque:
        queue = deque(maxlen=queue_size)
        on_msg = functools.partial(cls._on_message, queue)
        _ = ETopicListener(
            topic=topic,
            on_message=on_msg,
            use_bot_prefix=use_bot_prefix
        )
        return queue

    @classmethod
    def _on_message(cls, queue: deque, msg: Dict[str, Any], topic: str):
        queue.append((topic, msg))


class ExternalEventFactory:
    @classmethod
    def create_queue(cls,
                     event_name: str,
                     queue_size: Optional[int] = 1000
                     ) -> deque:
        return EEventQueueFactory.create(event_name, queue_size)

    @classmethod
    def create_async(cls,
                     event_name: str,
                     callback: Callable[[Dict[str, Any], str], None],
                     ) -> None:
        return EEventListenerFactory.create(event_name, callback)

    @classmethod
    def remove_listener(cls,
                        event_name: str,
                        callback: Callable[[Dict[str, Any], str], None],
                        ) -> None:
        EEventListenerFactory.remove(event_name, callback)


class ExternalTopicFactory:
    @classmethod
    def create_queue(cls,
                     topic: str,
                     queue_size: Optional[int] = 1000,
                     use_bot_prefix: Optional[bool] = True
                     ) -> deque:
        return ETopicQueueFactory.create(topic, queue_size, use_bot_prefix)

    @classmethod
    def create_async(cls,
                     topic: str,
                     callback: Callable[[Dict[str, Any], str], None],
                     use_bot_prefix: Optional[bool] = True
                     ) -> ETopicListener:
        return ETopicListenerFactory.create(topic, callback, use_bot_prefix)

    @classmethod
    def remove_listener(cls, listener):
        return ETopicListenerFactory.remove(listener)


class ETopicPublisher:
    def __init__(self,
                 topic: str,
                 use_bot_prefix: Optional[bool] = False):
        self._node = MQTTGateway.main()
        if self._node is None:
            raise Exception('MQTT Gateway not yet initialized')
        self._topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._node._hb_app.instance_id
        )
        if use_bot_prefix:
            self._topic = f'{self._topic_prefix}/{topic}'
        else:
            self._topic = topic
        self._pub = self._node.create_mpublisher()
        if self._node.state == NodeState.RUNNING:
            self._pub.run()

    def send(self, msg: Dict[str, Any]):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            asyncio.get_event_loop().call_soon_threadsafe(self.send, msg)
            return
        self._pub.publish(msg, self._topic)

    def __call__(self, msg: Dict[str, Any]):
        self.send(msg)


class EMTopicPublisher:
    def __init__(self,
                 use_bot_prefix: Optional[bool] = False):
        self._use_bot_prefix = use_bot_prefix
        self._node = MQTTGateway.main()
        if self._node is None:
            raise Exception('MQTT Gateway not yet initialized')
        self._topic_prefix = TopicSpecs.PREFIX.format(
            namespace=self._node.namespace,
            instance_id=self._node._hb_app.instance_id
        )

        self._pub = self._node.create_mpublisher()
        if self._node.state == NodeState.RUNNING:
            self._pub.run()

    def send(self, topic: str, msg: Dict[str, Any]):
        if threading.current_thread() != threading.main_thread():  # pragma: no cover
            asyncio.get_event_loop().call_soon_threadsafe(self.send, msg)
            return
        _topic = self._make_topic(topic)
        self._pub.publish(msg, _topic)

    def _make_topic(self, topic: str):
        if self._use_bot_prefix:
            _topic = f'{self._topic_prefix}/{topic}'
        else:
            _topic = topic
        return _topic

    def __call__(self, topic: str, msg: Dict[str, Any]):
        self.send(topic, msg)



from decimal import Decimal
from typing import Optional

from pydantic import ConfigDict, SecretStr, field_validator
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# Maker rebates(-0.02%) are paid out continuously on each trade directly to the trading wallet.(https://hyperliquid.gitbook.io/hyperliquid-docs/trading/fees)
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0"),
    taker_percent_fee_decimal=Decimal("0.00025"),
    buy_percent_fee_deducted_from_returns=True
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USD"

BROKER_ID = "HBOT"


def validate_bool(value: str) -> Optional[str]:
    """
    Permissively interpret a string as a boolean
    """
    valid_values = ('true', 'yes', 'y', 'false', 'no', 'n')
    if value.lower() not in valid_values:
        return f"Invalid value, please choose value from {valid_values}"


class HyperliquidPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="hyperliquid_perpetual", client_data=None)
    hyperliquid_perpetual_api_secret: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Arbitrum wallet private key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    use_vault: bool = Field(
        default="no",
        client_data=ClientFieldData(
            prompt=lambda cm: "Do you want to use the vault address?(Yes/No)",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )
    hyperliquid_perpetual_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Arbitrum or vault address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )

    @field_validator("use_vault", mode="before")
    @classmethod
    def validate_bool(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_bool(v)
            if ret is not None:
                raise ValueError(ret)
        return v


KEYS = HyperliquidPerpetualConfigMap.construct()

OTHER_DOMAINS = ["hyperliquid_perpetual_testnet"]
OTHER_DOMAINS_PARAMETER = {"hyperliquid_perpetual_testnet": "hyperliquid_perpetual_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"hyperliquid_perpetual_testnet": "BTC-USD"}
OTHER_DOMAINS_DEFAULT_FEES = {"hyperliquid_perpetual_testnet": [0, 0.025]}


class HyperliquidPerpetualTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="hyperliquid_perpetual_testnet", client_data=None)
    hyperliquid_perpetual_testnet_api_secret: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Arbitrum wallet private key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    use_vault: bool = Field(
        default="no",
        client_data=ClientFieldData(
            prompt=lambda cm: "Do you want to use the vault address?(Yes/No)",
            is_secure=False,
            is_connect_key=True,
            prompt_on_new=True,
        ),
    )
    hyperliquid_perpetual_testnet_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Arbitrum or vault address",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="hyperliquid_perpetual")

    @field_validator("use_vault", mode="before")
    @classmethod
    def validate_bool(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_bool(v)
            if ret is not None:
                raise ValueError(ret)
        return v


OTHER_DOMAINS_KEYS = {"hyperliquid_perpetual_testnet": HyperliquidPerpetualTestnetConfigMap.construct()}



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.vertex import (
    vertex_constants as CONSTANTS,
    vertex_utils as utils,
    vertex_web_utils as web_utils,
)
from hummingbot.connector.exchange.vertex.vertex_auth import VertexAuth
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.exchange.vertex.vertex_exchange import VertexExchange


class VertexAPIUserStreamDataSource(UserStreamTrackerDataSource):
    def __init__(
        self,
        auth: VertexAuth,
        trading_pairs: List[str],
        connector: "VertexExchange",
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
        api_factory: Optional[WebAssistantsFactory] = None,
        throttler: Optional[AsyncThrottler] = None,
    ):
        super().__init__()
        self._connector = connector
        self._auth: VertexAuth = auth
        self._trading_pairs = trading_pairs
        self._last_recv_time: float = 0
        self._domain = domain
        self._throttler = throttler
        self._api_factory = api_factory or web_utils.build_api_factory(throttler=self._throttler, auth=self._auth)
        self._ping_interval = 0
        self._last_ws_message_sent_timestamp = 0

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws_url = f"{CONSTANTS.WS_SUBSCRIBE_URLS[self._domain]}"
        self._ping_interval = CONSTANTS.HEARTBEAT_TIME_INTERVAL

        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=ws_url, message_timeout=self._ping_interval)
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                product_id = utils.trading_pair_to_product_id(
                    trading_pair, self._connector._exchange_market_info[self._domain]
                )

                fill_payload = {
                    "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                    "stream": {
                        "type": CONSTANTS.FILL_EVENT_TYPE,
                        "product_id": product_id,
                        "subaccount": self._auth.sender_address,
                    },
                    "id": product_id,
                }
                position_change_payload = {
                    "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                    "stream": {
                        "type": CONSTANTS.POSITION_CHANGE_EVENT_TYPE,
                        "product_id": product_id,
                        "subaccount": self._auth.sender_address,
                    },
                    "id": product_id,
                }

                subscribe_fill_request: WSJSONRequest = WSJSONRequest(payload=fill_payload)
                subscribe_position_change_request: WSJSONRequest = WSJSONRequest(payload=position_change_payload)
                await websocket_assistant.send(subscribe_fill_request)
                await websocket_assistant.send(subscribe_position_change_request)

                self._last_ws_message_sent_timestamp = self._time()

                self.logger().info(f"Subscribed to subaccount fill and position change channels of {trading_pair}...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to trading and order book stream...", exc_info=True
            )
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                seconds_until_next_ping = self._ping_interval - (self._time() - self._last_ws_message_sent_timestamp)
                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant, queue=queue),
                    timeout=seconds_until_next_ping,
                )
            except asyncio.TimeoutError:
                ping_time = self._time()
                await websocket_assistant.ping()
                self._last_ws_message_sent_timestamp = ping_time

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if (
            len(event_message) > 0
            and "type" in event_message
            and event_message.get("type") in [CONSTANTS.POSITION_CHANGE_EVENT_TYPE, CONSTANTS.FILL_EVENT_TYPE]
        ):
            queue.put_nowait(event_message)



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional


class Chain(Enum):
    ETHEREUM = ('ethereum', 'ETH')
    SOLANA = ('solana', 'SOL')

    def __init__(self, chain: str, native_currency: str):
        self.chain = chain
        self.native_currency = native_currency


class Connector(Enum):
    def __int__(self, chain: Chain, connector: str):
        self.chain = chain
        self.connector = connector


@dataclass
class PlaceOrderResult:
    update_timestamp: float
    client_order_id: str
    exchange_order_id: Optional[str]
    trading_pair: str
    misc_updates: Dict[str, Any] = field(default_factory=lambda: {})
    exception: Optional[Exception] = None


@dataclass
class CancelOrderResult:
    client_order_id: str
    trading_pair: str
    misc_updates: Dict[str, Any] = field(default_factory=lambda: {})
    not_found: bool = False
    exception: Optional[Exception] = None



import asyncio
import time
from json import JSONDecodeError
from typing import Any, Dict, Mapping, Optional

import aiohttp
from aiohttp import WebSocketError, WSCloseCode

from hummingbot.core.web_assistant.connections.data_types import WSRequest, WSResponse


class WSConnection:
    _MAX_MSG_SIZE = 4 * 1024 * 1024  # default aiohttp: 4 * 1024 * 1024

    def __init__(self, aiohttp_client_session: aiohttp.ClientSession):
        self._client_session = aiohttp_client_session
        self._connection: Optional[aiohttp.ClientWebSocketResponse] = None
        self._connected = False
        self._message_timeout: Optional[float] = None
        self._last_recv_time = 0

    @property
    def last_recv_time(self) -> float:
        return self._last_recv_time

    @property
    def connected(self) -> bool:
        return self._connected

    async def connect(
        self,
        ws_url: str,
        ping_timeout: float = 10,
        message_timeout: Optional[float] = None,
        ws_headers: Optional[Dict] = {},
        max_msg_size: Optional[int] = None
    ):
        self._ensure_not_connected()
        self._connection = await self._client_session.ws_connect(
            ws_url,
            headers=ws_headers,
            autoping=False,
            heartbeat=ping_timeout,
            max_msg_size=max_msg_size,
        )
        self._message_timeout = message_timeout
        self._connected = True

    async def disconnect(self):
        if self._connection is not None and not self._connection.closed:
            await self._connection.close()
        self._connection = None
        self._connected = False

    async def send(self, request: WSRequest):
        self._ensure_connected()
        await request.send_with_connection(connection=self)

    async def ping(self):
        await self._connection.ping()

    async def receive(self) -> Optional[WSResponse]:
        self._ensure_connected()
        response = None
        while self._connected:
            msg = await self._read_message()
            msg = await self._process_message(msg)
            if msg is not None:
                response = self._build_resp(msg)
                break
        return response

    def _ensure_not_connected(self):
        if self._connected:
            raise RuntimeError("WS is connected.")

    def _ensure_connected(self):
        if not self._connected:
            raise RuntimeError("WS is not connected.")

    async def _read_message(self) -> aiohttp.WSMessage:
        try:
            msg = await self._connection.receive(self._message_timeout)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError("Message receive timed out.")
        return msg

    async def _process_message(self, msg: aiohttp.WSMessage) -> Optional[aiohttp.WSMessage]:
        msg = await self._check_msg_types(msg)
        self._update_last_recv_time(msg)
        return msg

    async def _check_msg_types(self, msg: aiohttp.WSMessage) -> Optional[aiohttp.WSMessage]:
        msg = await self._check_msg_too_big_type(msg)
        msg = await self._check_msg_closed_type(msg)
        msg = await self._check_msg_ping_type(msg)
        msg = await self._check_msg_pong_type(msg)
        return msg

    async def _check_msg_too_big_type(self, msg: Optional[aiohttp.WSMessage]) -> Optional[aiohttp.WSMessage]:
        if msg is not None and msg.type in [aiohttp.WSMsgType.ERROR]:
            if isinstance(msg.data, WebSocketError) and msg.data.code == WSCloseCode.MESSAGE_TOO_BIG:
                await self.disconnect()
                raise WebSocketError(message=f"The WS message is too big: {msg.data}", code=WSCloseCode.MESSAGE_TOO_BIG)
            else:
                await self.disconnect()
                raise ConnectionError(f"WS error: {msg.data}")
        return msg

    async def _check_msg_closed_type(self, msg: Optional[aiohttp.WSMessage]) -> Optional[aiohttp.WSMessage]:
        if msg is not None and msg.type in [aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE]:
            if self._connected:
                close_code = self._connection.close_code
                await self.disconnect()
                raise ConnectionError(
                    f"The WS connection was closed unexpectedly. Close code = {close_code} msg data: {msg.data}"
                )
            msg = None
        return msg

    async def _check_msg_ping_type(self, msg: Optional[aiohttp.WSMessage]) -> Optional[aiohttp.WSMessage]:
        if msg is not None and msg.type == aiohttp.WSMsgType.PING:
            await self._connection.pong(msg.data)
            msg = None
        return msg

    async def _check_msg_pong_type(self, msg: Optional[aiohttp.WSMessage]) -> Optional[aiohttp.WSMessage]:
        if msg is not None and msg.type == aiohttp.WSMsgType.PONG:
            msg = None
        return msg

    def _update_last_recv_time(self, _: aiohttp.WSMessage):
        self._last_recv_time = time.time()

    async def _send_json(self, payload: Mapping[str, Any]):
        await self._connection.send_json(payload)

    async def _send_plain_text(self, payload: str):
        await self._connection.send_str(payload)

    async def _send_binary(self, payload: bytes):
        await self._connection.send_bytes(payload)

    @staticmethod
    def _build_resp(msg: aiohttp.WSMessage) -> WSResponse:
        if msg.type == aiohttp.WSMsgType.BINARY:
            data = msg.data
        else:
            try:
                data = msg.json()
            except JSONDecodeError:
                data = msg.data
        response = WSResponse(data)
        return response



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional

from grpc import RpcError
from pyinjective.async_client import AsyncClient
from pyinjective.client.model.pagination import PaginationOption
from pyinjective.core.market import DerivativeMarket, SpotMarket
from pyinjective.core.token import Token
from pyinjective.proto.injective.stream.v1beta1 import query_pb2 as chain_stream_query


class BaseInjectiveQueryExecutor(ABC):

    @abstractmethod
    async def ping(self):  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def spot_markets(self) -> Dict[str, SpotMarket]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def derivative_markets(self) -> Dict[str, DerivativeMarket]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def tokens(self) -> Dict[str, Token]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def derivative_market(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_spot_orderbook(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_derivative_orderbook(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_tx(self, tx_hash: str) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def account_portfolio(self, account_address: str) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def simulate_tx(self, tx_byte: bytes) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def send_tx_sync_mode(self, tx_byte: bytes) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_spot_trades(
            self,
            market_ids: List[str],
            subaccount_id: Optional[str] = None,
            start_time: Optional[int] = None,
            skip: Optional[int] = None,
            limit: Optional[int] = None,
    ) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_derivative_trades(
            self,
            market_ids: List[str],
            subaccount_id: Optional[str] = None,
            start_time: Optional[int] = None,
            skip: Optional[int] = None,
            limit: Optional[int] = None,
    ) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_historical_spot_orders(
            self,
            market_ids: List[str],
            subaccount_id: str,
            start_time: int,
            skip: int,
    ) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_historical_derivative_orders(
            self,
            market_ids: List[str],
            subaccount_id: str,
            start_time: int,
            skip: int,
    ) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_funding_rates(self, market_id: str, limit: int) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_oracle_prices(
            self,
            base_symbol: str,
            quote_symbol: str,
            oracle_type: str,
            oracle_scale_factor: int,
    ) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_funding_payments(self, subaccount_id: str, market_id: str, limit: int) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def get_derivative_positions(self, subaccount_id: str, skip: int) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def listen_transactions_updates(
        self,
        callback: Callable,
        on_end_callback: Callable,
        on_status_callback: Callable,
    ):  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    async def listen_chain_stream_updates(
        self,
        callback: Callable,
        on_end_callback: Callable,
        on_status_callback: Callable,
        bank_balances_filter: Optional[chain_stream_query.BankBalancesFilter] = None,
        subaccount_deposits_filter: Optional[chain_stream_query.SubaccountDepositsFilter] = None,
        spot_trades_filter: Optional[chain_stream_query.TradesFilter] = None,
        derivative_trades_filter: Optional[chain_stream_query.TradesFilter] = None,
        spot_orders_filter: Optional[chain_stream_query.OrdersFilter] = None,
        derivative_orders_filter: Optional[chain_stream_query.OrdersFilter] = None,
        spot_orderbooks_filter: Optional[chain_stream_query.OrderbookFilter] = None,
        derivative_orderbooks_filter: Optional[chain_stream_query.OrderbookFilter] = None,
        positions_filter: Optional[chain_stream_query.PositionsFilter] = None,
        oracle_price_filter: Optional[chain_stream_query.OraclePriceFilter] = None,
    ):
        raise NotImplementedError


class PythonSDKInjectiveQueryExecutor(BaseInjectiveQueryExecutor):

    def __init__(self, sdk_client: AsyncClient):
        super().__init__()
        self._sdk_client = sdk_client

    async def ping(self):  # pragma: no cover
        await self._sdk_client.fetch_ping()

    async def spot_markets(self) -> Dict[str, SpotMarket]:  # pragma: no cover
        return await self._sdk_client.all_spot_markets()

    async def derivative_markets(self) -> Dict[str, DerivativeMarket]:  # pragma: no cover
        return await self._sdk_client.all_derivative_markets()

    async def tokens(self) -> Dict[str, Token]:  # pragma: no cover
        return await self._sdk_client.all_tokens()

    async def derivative_market(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        response = await self._sdk_client.fetch_derivative_market(market_id=market_id)
        return response

    async def get_spot_orderbook(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        order_book_response = await self._sdk_client.fetch_spot_orderbook_v2(market_id=market_id)
        order_book_data = order_book_response["orderbook"]
        result = {
            "buys": [(buy["price"], buy["quantity"], int(buy["timestamp"])) for buy in order_book_data.get("buys", [])],
            "sells": [(sell["price"], sell["quantity"], int(sell["timestamp"])) for sell in order_book_data.get("sells", [])],
            "sequence": int(order_book_data["sequence"]),
            "timestamp": int(order_book_data["timestamp"]),
        }

        return result

    async def get_derivative_orderbook(self, market_id: str) -> Dict[str, Any]:  # pragma: no cover
        order_book_response = await self._sdk_client.fetch_derivative_orderbooks_v2(market_ids=[market_id])
        order_book_data = order_book_response["orderbooks"][0]["orderbook"]
        result = {
            "buys": [(buy["price"], buy["quantity"], int(buy["timestamp"])) for buy in order_book_data.get("buys", [])],
            "sells": [(sell["price"], sell["quantity"], int(sell["timestamp"])) for sell in
                      order_book_data.get("sells", [])],
            "sequence": int(order_book_data["sequence"]),
            "timestamp": int(order_book_data["timestamp"]),
        }

        return result

    async def get_tx(self, tx_hash: str) -> Dict[str, Any]:  # pragma: no cover
        try:
            transaction_response = await self._sdk_client.fetch_tx(hash=tx_hash)
        except RpcError as rpc_exception:
            if "tx not found" in str(rpc_exception):
                raise ValueError(f"The transaction with hash {tx_hash} was not found")
            else:
                raise

        return transaction_response

    async def account_portfolio(self, account_address: str) -> Dict[str, Any]:  # pragma: no cover
        portfolio_response = await self._sdk_client.fetch_account_portfolio_balances(account_address=account_address)
        return portfolio_response

    async def simulate_tx(self, tx_byte: bytes) -> Dict[str, Any]:  # pragma: no cover
        try:
            response = await self._sdk_client.simulate(tx_bytes=tx_byte)
        except RpcError as ex:
            raise RuntimeError(f"Transaction simulation failure ({ex})")
        return response

    async def send_tx_sync_mode(self, tx_byte: bytes) -> Dict[str, Any]:  # pragma: no cover
        response = await self._sdk_client.broadcast_tx_sync_mode(tx_bytes=tx_byte)
        result = response["txResponse"]
        return result

    async def get_spot_trades(
            self,
            market_ids: List[str],
            subaccount_id: Optional[str] = None,
            start_time: Optional[int] = None,
            skip: Optional[int] = None,
            limit: Optional[int] = None,
    ) -> Dict[str, Any]:  # pragma: no cover
        subaccount_ids = [subaccount_id] if subaccount_id is not None else None
        pagination = PaginationOption(skip=skip, limit=limit, start_time=start_time)
        response = await self._sdk_client.fetch_spot_trades(
            market_ids=market_ids,
            subaccount_ids=subaccount_ids,
            pagination=pagination,
        )
        return response

    async def get_derivative_trades(
            self,
            market_ids: List[str],
            subaccount_id: Optional[str] = None,
            start_time: Optional[int] = None,
            skip: Optional[int] = None,
            limit: Optional[int] = None,
    ) -> Dict[str, Any]:  # pragma: no cover
        subaccount_ids = [subaccount_id] if subaccount_id is not None else None
        pagination = PaginationOption(skip=skip, limit=limit, start_time=start_time)
        response = await self._sdk_client.fetch_derivative_trades(
            market_ids=market_ids,
            subaccount_ids=subaccount_ids,
            pagination=pagination,
        )
        return response

    async def get_historical_spot_orders(
            self,
            market_ids: List[str],
            subaccount_id: str,
            start_time: int,
            skip: int,
    ) -> Dict[str, Any]:  # pragma: no cover
        pagination = PaginationOption(skip=skip, start_time=start_time)
        response = await self._sdk_client.fetch_spot_orders_history(
            market_ids=market_ids,
            subaccount_id=subaccount_id,
            pagination=pagination
        )
        return response

    async def get_historical_derivative_orders(
            self,
            market_ids: List[str],
            subaccount_id: str,
            start_time: int,
            skip: int,
    ) -> Dict[str, Any]:  # pragma: no cover
        pagination = PaginationOption(skip=skip, start_time=start_time)
        response = await self._sdk_client.fetch_derivative_orders_history(
            market_ids=market_ids,
            subaccount_id=subaccount_id,
            pagination=pagination,
        )
        return response

    async def get_funding_rates(self, market_id: str, limit: int) -> Dict[str, Any]:  # pragma: no cover
        pagination = PaginationOption(limit=limit)
        response = await self._sdk_client.fetch_funding_rates(market_id=market_id, pagination=pagination)
        return response

    async def get_funding_payments(self, subaccount_id: str, market_id: str, limit: int) -> Dict[str, Any]:    # pragma: no cover
        pagination = PaginationOption(limit=limit)
        response = await self._sdk_client.fetch_funding_payments(
            market_ids=[market_id],
            subaccount_id=subaccount_id,
            pagination=pagination,
        )
        return response

    async def get_derivative_positions(self, subaccount_id: str, skip: int) -> Dict[str, Any]:    # pragma: no cover
        pagination = PaginationOption(skip=skip)
        response = await self._sdk_client.fetch_derivative_positions_v2(
            subaccount_id=subaccount_id, pagination=pagination
        )
        return response

    async def get_oracle_prices(
            self,
            base_symbol: str,
            quote_symbol: str,
            oracle_type: str,
            oracle_scale_factor: int,
    ) -> Dict[str, Any]:    # pragma: no cover
        response = await self._sdk_client.fetch_oracle_price(
            base_symbol=base_symbol,
            quote_symbol=quote_symbol,
            oracle_type=oracle_type,
            oracle_scale_factor=oracle_scale_factor
        )
        return response

    async def listen_transactions_updates(
        self,
        callback: Callable,
        on_end_callback: Callable,
        on_status_callback: Callable,
    ):  # pragma: no cover
        await self._sdk_client.listen_txs_updates(
            callback=callback,
            on_end_callback=on_end_callback,
            on_status_callback=on_status_callback,
        )

    async def listen_chain_stream_updates(
        self,
        callback: Callable,
        on_end_callback: Callable,
        on_status_callback: Callable,
        bank_balances_filter: Optional[chain_stream_query.BankBalancesFilter] = None,
        subaccount_deposits_filter: Optional[chain_stream_query.SubaccountDepositsFilter] = None,
        spot_trades_filter: Optional[chain_stream_query.TradesFilter] = None,
        derivative_trades_filter: Optional[chain_stream_query.TradesFilter] = None,
        spot_orders_filter: Optional[chain_stream_query.OrdersFilter] = None,
        derivative_orders_filter: Optional[chain_stream_query.OrdersFilter] = None,
        spot_orderbooks_filter: Optional[chain_stream_query.OrderbookFilter] = None,
        derivative_orderbooks_filter: Optional[chain_stream_query.OrderbookFilter] = None,
        positions_filter: Optional[chain_stream_query.PositionsFilter] = None,
        oracle_price_filter: Optional[chain_stream_query.OraclePriceFilter] = None,
    ):  # pragma: no cover
        await self._sdk_client.listen_chain_stream_updates(
            callback=callback,
            on_end_callback=on_end_callback,
            on_status_callback=on_status_callback,
            bank_balances_filter=bank_balances_filter,
            subaccount_deposits_filter=subaccount_deposits_filter,
            spot_trades_filter=spot_trades_filter,
            derivative_trades_filter=derivative_trades_filter,
            spot_orders_filter=spot_orders_filter,
            derivative_orders_filter=derivative_orders_filter,
            spot_orderbooks_filter=spot_orderbooks_filter,
            derivative_orderbooks_filter=derivative_orderbooks_filter,
            positions_filter=positions_filter,
            oracle_price_filter=oracle_price_filter,
        )



from typing import TYPE_CHECKING, Optional

from hummingbot.client.config.config_helpers import parse_config_default_to_text, parse_cvar_value
from hummingbot.client.config.config_validators import validate_bool
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.core.utils.async_utils import safe_ensure_future

from .import_command import ImportCommand

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication


class PreviousCommand:
    def previous_strategy(
        self,  # type: HummingbotApplication
        option: str,
    ):
        if option is not None:
            pass

        previous_strategy_file = self.client_config_map.previous_strategy

        if previous_strategy_file is not None:
            safe_ensure_future(self.prompt_for_previous_strategy(previous_strategy_file))
        else:
            self.notify("No previous strategy found.")

    async def prompt_for_previous_strategy(
        self,  # type: HummingbotApplication
        file_name: str,
    ):
        self.app.clear_input()
        self.placeholder_mode = True
        self.app.hide_input = True

        previous_strategy = ConfigVar(
            key="previous_strategy_answer",
            prompt=f"Do you want to import the previously stored config? [{file_name}] (Yes/No) >>>",
            type_str="bool",
            validator=validate_bool,
        )

        await self.prompt_answer(previous_strategy)
        if self.app.to_stop_config:
            self.app.to_stop_config = False
            return

        if previous_strategy.value:
            ImportCommand.import_command(self, file_name)

        # clean
        self.app.change_prompt(prompt=">>> ")

        # reset input
        self.placeholder_mode = False
        self.app.hide_input = False

    async def prompt_answer(
        self,  # type: HummingbotApplication
        config: ConfigVar,
        input_value: Optional[str] = None,
        assign_default: bool = True,
    ):

        if input_value is None:
            if assign_default:
                self.app.set_text(parse_config_default_to_text(config))
            prompt = await config.get_prompt()
            input_value = await self.app.prompt(prompt=prompt)

        if self.app.to_stop_config:
            return
        config.value = parse_cvar_value(config, input_value)
        err_msg = await config.validate(input_value)
        if err_msg is not None:
            self.notify(err_msg)
            config.value = None
            await self.prompt_answer(config)



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Mapping, Optional

import aiohttp
import ujson

if TYPE_CHECKING:
    from hummingbot.core.web_assistant.connections.ws_connection import WSConnection


class RESTMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"

    def __str__(self):
        obj_str = repr(self)
        return obj_str

    def __repr__(self):
        return self.value


@dataclass
class RESTRequest:
    method: RESTMethod
    url: Optional[str] = None
    endpoint_url: Optional[str] = None
    params: Optional[Mapping[str, str]] = None
    data: Any = None
    headers: Optional[Mapping[str, str]] = None
    is_auth_required: bool = False
    throttler_limit_id: Optional[str] = None


@dataclass
class EndpointRESTRequest(RESTRequest, ABC):
    """This request class enable the user to provide either a complete URL or simply an endpoint.

    The endpoint is concatenated with the return value of `base_url`. It can handle endpoints supplied both as
    `"endpoint"` and `"/endpoint"`. It also provides the necessary checks to ensure a valid URL can be constructed.
    """

    endpoint: Optional[str] = None

    def __post_init__(self):
        self._ensure_url()
        self._ensure_params()
        self._ensure_data()

    @property
    @abstractmethod
    def base_url(self) -> str:
        ...

    def _ensure_url(self):
        if self.url is None and self.endpoint is None:
            raise ValueError("Either the full url or the endpoint must be specified.")
        if self.url is None:
            if self.endpoint.startswith("/"):
                self.url = f"{self.base_url}{self.endpoint}"
            else:
                self.url = f"{self.base_url}/{self.endpoint}"

    def _ensure_params(self):
        if self.method == RESTMethod.POST:
            if self.params is not None:
                raise ValueError("POST requests should not use `params`. Use `data` instead.")

    def _ensure_data(self):
        if self.method == RESTMethod.POST:
            if self.data is not None:
                self.data = ujson.dumps(self.data)
        elif self.data is not None:
            raise ValueError("The `data` field should be used only for POST requests. Use `params` instead.")


@dataclass(init=False)
class RESTResponse:
    url: str
    method: RESTMethod
    status: int
    headers: Optional[Mapping[str, str]]

    def __init__(self, aiohttp_response: aiohttp.ClientResponse):
        self._aiohttp_response = aiohttp_response

    @property
    def url(self) -> str:
        url_str = str(self._aiohttp_response.url)
        return url_str

    @property
    def method(self) -> RESTMethod:
        method_ = RESTMethod[self._aiohttp_response.method.upper()]
        return method_

    @property
    def status(self) -> int:
        status_ = int(self._aiohttp_response.status)
        return status_

    @property
    def headers(self) -> Optional[Mapping[str, str]]:
        headers_ = self._aiohttp_response.headers
        return headers_

    async def json(self) -> Any:
        if self._aiohttp_response.content_type == "text/html":
            byte_string = await self._aiohttp_response.read()
            if isinstance(byte_string, bytes):
                decoded_string = byte_string.decode('utf-8')
                json_ = json.loads(decoded_string)
            else:
                json_ = await self._aiohttp_response.json()
        else:
            json_ = await self._aiohttp_response.json()
        return json_

    async def text(self) -> str:
        text_ = await self._aiohttp_response.text()
        return text_


class WSRequest(ABC):
    @abstractmethod
    async def send_with_connection(self, connection: "WSConnection"):
        return NotImplemented


@dataclass
class WSJSONRequest(WSRequest):
    payload: Mapping[str, Any]
    throttler_limit_id: Optional[str] = None
    is_auth_required: bool = False

    async def send_with_connection(self, connection: "WSConnection"):
        await connection._send_json(payload=self.payload)


@dataclass
class WSPlainTextRequest(WSRequest):
    payload: str
    throttler_limit_id: Optional[str] = None
    is_auth_required: bool = False

    async def send_with_connection(self, connection: "WSConnection"):
        await connection._send_plain_text(payload=self.payload)


@dataclass
class WSBinaryRequest(WSRequest):
    payload: bytes
    throttler_limit_id: Optional[str] = None
    is_auth_required: bool = False

    async def send_with_connection(self, connection: "WSConnection"):
        await connection._send_binary(payload=self.payload)


@dataclass
class WSResponse:
    data: Any



import asyncio
import json
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

from google.protobuf import any_pb2, json_format
from pyinjective import Transaction
from pyinjective.async_client import AsyncClient
from pyinjective.composer import Composer, injective_exchange_tx_pb
from pyinjective.core.network import Network
from pyinjective.wallet import Address, PrivateKey

from hummingbot.connector.exchange.injective_v2 import injective_constants as CONSTANTS
from hummingbot.connector.exchange.injective_v2.data_sources.injective_data_source import InjectiveDataSource
from hummingbot.connector.exchange.injective_v2.injective_market import (
    InjectiveDerivativeMarket,
    InjectiveSpotMarket,
    InjectiveToken,
)
from hummingbot.connector.exchange.injective_v2.injective_query_executor import PythonSDKInjectiveQueryExecutor
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder, GatewayPerpetualInFlightOrder
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.api_throttler.async_throttler_base import AsyncThrottlerBase
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, TradeType
from hummingbot.core.data_type.in_flight_order import OrderState, OrderUpdate
from hummingbot.core.pubsub import PubSub
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.injective_v2.injective_v2_utils import InjectiveFeeCalculatorMode


class InjectiveVaultsDataSource(InjectiveDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            private_key: str,
            subaccount_index: int,
            vault_contract_address: str,
            vault_subaccount_index: int,
            network: Network,
            rate_limits: List[RateLimit],
            fee_calculator_mode: "InjectiveFeeCalculatorMode",
            use_secure_connection: bool = True):
        self._network = network
        self._client = AsyncClient(
            network=self._network,
            # insecure=not use_secure_connection,
        )
        self._composer = None
        self._query_executor = PythonSDKInjectiveQueryExecutor(sdk_client=self._client)
        self._fee_calculator_mode = fee_calculator_mode
        self._fee_calculator = None

        self._private_key = None
        self._public_key = None
        self._vault_admin_address = ""
        self._vault_admin_subaccount_index = subaccount_index
        self._vault_admin_subaccount_id = ""
        if private_key:
            self._private_key = PrivateKey.from_hex(private_key)
            self._public_key = self._private_key.to_public_key()
            self._vault_admin_address = self._public_key.to_address()
            self._vault_admin_subaccount_id = self._vault_admin_address.get_subaccount_id(index=subaccount_index)

        self._vault_contract_address = None
        self._vault_subaccount_id = ""
        self._vault_subaccount_index = vault_subaccount_index
        if vault_contract_address:
            self._vault_contract_address = Address.from_acc_bech32(vault_contract_address)
            self._vault_subaccount_id = self._vault_contract_address.get_subaccount_id(index=vault_subaccount_index)

        self._publisher = PubSub()
        self._last_received_message_timestamp = 0
        self._throttler = AsyncThrottler(rate_limits=rate_limits)

        self._is_timeout_height_initialized = False
        self._is_trading_account_initialized = False
        self._markets_initialization_lock = asyncio.Lock()
        self._spot_market_info_map: Optional[Dict[str, InjectiveSpotMarket]] = None
        self._derivative_market_info_map: Optional[Dict[str, InjectiveDerivativeMarket]] = None
        self._spot_market_and_trading_pair_map: Optional[Mapping[str, str]] = None
        self._derivative_market_and_trading_pair_map: Optional[Mapping[str, str]] = None
        self._tokens_map: Optional[Dict[str, InjectiveToken]] = None
        self._token_symbol_and_denom_map: Optional[Mapping[str, str]] = None

        self._events_listening_tasks: List[asyncio.Task] = []

    @property
    def publisher(self):
        return self._publisher

    @property
    def query_executor(self):
        return self._query_executor

    @property
    def throttler(self):
        return self._throttler

    @property
    def portfolio_account_injective_address(self) -> str:
        return self._vault_contract_address.to_acc_bech32()

    @property
    def portfolio_account_subaccount_id(self) -> str:
        return self._vault_subaccount_id

    @property
    def trading_account_injective_address(self) -> str:
        return self._vault_admin_address.to_acc_bech32()

    @property
    def injective_chain_id(self) -> str:
        return self._network.chain_id

    @property
    def fee_denom(self) -> str:
        return self._network.fee_denom

    @property
    def portfolio_account_subaccount_index(self) -> int:
        return self._vault_subaccount_index

    @property
    def network_name(self) -> str:
        return self._network.string()

    @property
    def last_received_message_timestamp(self) -> float:
        return self._last_received_message_timestamp

    async def composer(self) -> Composer:
        if self._composer is None:
            self._composer = await self._client.composer()
        return self._composer

    def events_listening_tasks(self) -> List[asyncio.Task]:
        return self._events_listening_tasks.copy()

    def add_listening_task(self, task: asyncio.Task):
        self._events_listening_tasks.append(task)

    async def timeout_height(self) -> int:
        if not self._is_timeout_height_initialized:
            await self._initialize_timeout_height()
        return self._client.timeout_height

    async def spot_market_and_trading_pair_map(self):
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()
        return self._spot_market_and_trading_pair_map.copy()

    async def spot_market_info_for_id(self, market_id: str):
        if self._spot_market_info_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_info_map is None:
                    await self.update_markets()

        return self._spot_market_info_map[market_id]

    async def derivative_market_and_trading_pair_map(self):
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()
        return self._derivative_market_and_trading_pair_map.copy()

    async def derivative_market_info_for_id(self, market_id: str):
        if self._derivative_market_info_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_info_map is None:
                    await self.update_markets()

        return self._derivative_market_info_map[market_id]

    async def trading_pair_for_market(self, market_id: str):
        if self._spot_market_and_trading_pair_map is None or self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None or self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        trading_pair = self._spot_market_and_trading_pair_map.get(market_id)

        if trading_pair is None:
            trading_pair = self._derivative_market_and_trading_pair_map[market_id]
        return trading_pair

    async def market_id_for_spot_trading_pair(self, trading_pair: str) -> str:
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()

        return self._spot_market_and_trading_pair_map.inverse[trading_pair]

    async def market_id_for_derivative_trading_pair(self, trading_pair: str) -> str:
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        return self._derivative_market_and_trading_pair_map.inverse[trading_pair]

    async def spot_markets(self):
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()

        return list(self._spot_market_info_map.values())

    async def derivative_markets(self):
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        return list(self._derivative_market_info_map.values())

    async def token(self, denom: str) -> InjectiveToken:
        if self._tokens_map is None:
            async with self._markets_initialization_lock:
                if self._tokens_map is None:
                    await self.update_markets()

        return self._tokens_map.get(denom)

    def configure_throttler(self, throttler: AsyncThrottlerBase):
        self._throttler = throttler

    async def trading_account_sequence(self) -> int:
        if not self._is_trading_account_initialized:
            await self.initialize_trading_account()
        return self._client.get_sequence()

    async def trading_account_number(self) -> int:
        if not self._is_trading_account_initialized:
            await self.initialize_trading_account()
        return self._client.get_number()

    async def stop(self):
        await super().stop()
        self._events_listening_tasks = []

    async def initialize_trading_account(self):
        await self._client.fetch_account(address=self.trading_account_injective_address)
        self._is_trading_account_initialized = True

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER]

    async def update_markets(self):
        (
            self._tokens_map,
            self._token_symbol_and_denom_map,
            self._spot_market_info_map,
            self._spot_market_and_trading_pair_map,
            self._derivative_market_info_map,
            self._derivative_market_and_trading_pair_map,
        ) = await self._get_markets_and_tokens()

    async def order_updates_for_transaction(
            self,
            transaction_hash: str,
            spot_orders: Optional[List[GatewayInFlightOrder]] = None,
            perpetual_orders: Optional[List[GatewayPerpetualInFlightOrder]] = None,
    ) -> List[OrderUpdate]:
        spot_orders = spot_orders or []
        perpetual_orders = perpetual_orders or []
        order_updates = []

        async with self.throttler.execute_task(limit_id=CONSTANTS.GET_TRANSACTION_LIMIT_ID):
            transaction_info = await self.query_executor.get_tx(tx_hash=transaction_hash)

        if transaction_info["txResponse"]["code"] != CONSTANTS.TRANSACTION_SUCCEEDED_CODE:
            # The transaction failed. All orders should be marked as failed
            for order in (spot_orders + perpetual_orders):
                order_update = OrderUpdate(
                    trading_pair=order.trading_pair,
                    update_timestamp=self._time(),
                    new_state=OrderState.FAILED,
                    client_order_id=order.client_order_id,
                )
                order_updates.append(order_update)

        return order_updates

    def real_tokens_spot_trading_pair(self, unique_trading_pair: str) -> str:
        resulting_trading_pair = unique_trading_pair
        if (self._spot_market_and_trading_pair_map is not None
                and self._spot_market_info_map is not None):
            market_id = self._spot_market_and_trading_pair_map.inverse.get(unique_trading_pair)
            market = self._spot_market_info_map.get(market_id)
            if market is not None:
                resulting_trading_pair = combine_to_hb_trading_pair(
                    base=market.base_token.symbol,
                    quote=market.quote_token.symbol,
                )

        return resulting_trading_pair

    def real_tokens_perpetual_trading_pair(self, unique_trading_pair: str) -> str:
        resulting_trading_pair = unique_trading_pair
        if (self._derivative_market_and_trading_pair_map is not None
                and self._derivative_market_info_map is not None):
            market_id = self._derivative_market_and_trading_pair_map.inverse.get(unique_trading_pair)
            market = self._derivative_market_info_map.get(market_id)
            if market is not None:
                resulting_trading_pair = combine_to_hb_trading_pair(
                    base=market.base_token_symbol(),
                    quote=market.quote_token.symbol,
                )

        return resulting_trading_pair

    async def _initialize_timeout_height(self):
        await self._client.sync_timeout_height()
        self._is_timeout_height_initialized = True

    def _sign_and_encode(self, transaction: Transaction) -> bytes:
        sign_doc = transaction.get_sign_doc(self._public_key)
        sig = self._private_key.sign(sign_doc.SerializeToString())
        tx_raw_bytes = transaction.get_tx_data(sig, self._public_key)
        return tx_raw_bytes

    def _uses_default_portfolio_subaccount(self) -> bool:
        return self._vault_subaccount_index == CONSTANTS.DEFAULT_SUBACCOUNT_INDEX

    async def _updated_derivative_market_info_for_id(self, market_id: str) -> Dict[str, Any]:
        async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_MARKETS_LIMIT_ID):
            market_info = await self._query_executor.derivative_market(market_id=market_id)

        return market_info

    async def _order_creation_messages(
            self,
            spot_orders_to_create: List[GatewayInFlightOrder],
            derivative_orders_to_create: List[GatewayPerpetualInFlightOrder],
    ) -> List[any_pb2.Any]:
        composer = await self.composer()
        spot_order_definitions = []
        derivative_order_definitions = []

        for order in spot_orders_to_create:
            order_definition = await self._create_spot_order_definition(order=order)
            spot_order_definitions.append(order_definition)

        for order in derivative_orders_to_create:
            order_definition = await self._create_derivative_order_definition(order=order)
            derivative_order_definitions.append(order_definition)

        message = composer.msg_batch_update_orders(
            sender=self.portfolio_account_injective_address,
            spot_orders_to_create=spot_order_definitions,
            derivative_orders_to_create=derivative_order_definitions,
        )

        message_as_dictionary = json_format.MessageToDict(
            message=message,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )
        del message_as_dictionary["subaccount_id"]

        execute_message_parameter = self._create_execute_contract_internal_message(batch_update_orders_params=message_as_dictionary)

        execute_contract_message = composer.MsgExecuteContract(
            sender=self._vault_admin_address.to_acc_bech32(),
            contract=self._vault_contract_address.to_acc_bech32(),
            msg=json.dumps(execute_message_parameter),
        )

        return [execute_contract_message]

    async def _order_cancel_message(
            self,
            spot_orders_to_cancel: List[injective_exchange_tx_pb.OrderData],
            derivative_orders_to_cancel: List[injective_exchange_tx_pb.OrderData]
    ) -> any_pb2.Any:
        composer = await self.composer()

        message = composer.msg_batch_update_orders(
            sender=self.portfolio_account_injective_address,
            spot_orders_to_cancel=spot_orders_to_cancel,
            derivative_orders_to_cancel=derivative_orders_to_cancel,
        )

        message_as_dictionary = json_format.MessageToDict(
            message=message,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )
        del message_as_dictionary["subaccount_id"]

        execute_message_parameter = self._create_execute_contract_internal_message(batch_update_orders_params=message_as_dictionary)

        execute_contract_message = composer.MsgExecuteContract(
            sender=self._vault_admin_address.to_acc_bech32(),
            contract=self._vault_contract_address.to_acc_bech32(),
            msg=json.dumps(execute_message_parameter),
        )

        return execute_contract_message

    async def _all_subaccount_orders_cancel_message(
            self,
            spot_markets_ids: List[str],
            derivative_markets_ids: List[str]
    ) -> any_pb2.Any:
        composer = await self.composer()

        message = composer.msg_batch_update_orders(
            sender=self.portfolio_account_injective_address,
            subaccount_id=self.portfolio_account_subaccount_id,
            spot_market_ids_to_cancel_all=spot_markets_ids,
            derivative_market_ids_to_cancel_all=derivative_markets_ids,
        )

        message_as_dictionary = json_format.MessageToDict(
            message=message,
            always_print_fields_with_no_presence=True,
            preserving_proto_field_name=True,
            use_integers_for_enums=True,
        )

        execute_message_parameter = self._create_execute_contract_internal_message(
            batch_update_orders_params=message_as_dictionary)

        execute_contract_message = composer.MsgExecuteContract(
            sender=self._vault_admin_address.to_acc_bech32(),
            contract=self._vault_contract_address.to_acc_bech32(),
            msg=json.dumps(execute_message_parameter),
        )

        return execute_contract_message

    async def _generate_injective_order_data(self, order: GatewayInFlightOrder, market_id: str) -> injective_exchange_tx_pb.OrderData:
        composer = await self.composer()
        order_hash = order.exchange_order_id
        cid = order.client_order_id if order_hash is None else None
        order_data = composer.order_data(
            market_id=market_id,
            subaccount_id=str(self.portfolio_account_subaccount_index),
            order_hash=order_hash,
            cid=cid,
            is_buy=order.trade_type == TradeType.BUY,
            is_market_order=order.order_type == OrderType.MARKET,
        )

        return order_data

    async def _create_spot_order_definition(self, order: GatewayInFlightOrder):
        # Both price and quantity have to be adjusted because the vaults expect to receive those values without
        # the extra 18 zeros that the chain backend expects for direct trading messages
        order_type = "BUY" if order.trade_type == TradeType.BUY else "SELL"
        if order.order_type == OrderType.LIMIT_MAKER:
            order_type = order_type + "_PO"
        market_id = await self.market_id_for_spot_trading_pair(order.trading_pair)
        composer = await self.composer()
        definition = composer.spot_order(
            market_id=market_id,
            subaccount_id=str(self.portfolio_account_subaccount_index),
            fee_recipient=self.portfolio_account_injective_address,
            price=order.price,
            quantity=order.amount,
            order_type=order_type,
            cid=order.client_order_id,
        )

        definition.order_info.quantity = f"{(Decimal(definition.order_info.quantity) * Decimal('1e-18')).normalize():f}"
        definition.order_info.price = f"{(Decimal(definition.order_info.price) * Decimal('1e-18')).normalize():f}"
        return definition

    async def _create_derivative_order_definition(self, order: GatewayPerpetualInFlightOrder):
        # Price, quantity and margin have to be adjusted because the vaults expect to receive those values without
        # the extra 18 zeros that the chain backend expects for direct trading messages
        order_type = "BUY" if order.trade_type == TradeType.BUY else "SELL"
        if order.order_type == OrderType.LIMIT_MAKER:
            order_type = order_type + "_PO"
        market_id = await self.market_id_for_derivative_trading_pair(order.trading_pair)
        composer = await self.composer()
        definition = composer.derivative_order(
            market_id=market_id,
            subaccount_id=str(self.portfolio_account_subaccount_index),
            fee_recipient=self.portfolio_account_injective_address,
            price=order.price,
            quantity=order.amount,
            margin=composer.calculate_margin(
                quantity=order.amount,
                price=order.price,
                leverage=Decimal(str(order.leverage)),
                is_reduce_only=order.position == PositionAction.CLOSE,
            ),
            order_type=order_type,
            cid=order.client_order_id,
        )

        definition.order_info.quantity = f"{(Decimal(definition.order_info.quantity) * Decimal('1e-18')).normalize():f}"
        definition.order_info.price = f"{(Decimal(definition.order_info.price) * Decimal('1e-18')).normalize():f}"
        definition.margin = f"{(Decimal(definition.margin) * Decimal('1e-18')).normalize():f}"
        return definition

    def _create_execute_contract_internal_message(self, batch_update_orders_params: Dict) -> Dict[str, Any]:
        return {
            "admin_execute_message": {
                "injective_message": {
                    "custom": {
                        "route": "exchange",
                        "msg_data": {
                            "batch_update_orders": batch_update_orders_params
                        }
                    }
                }
            }
        }

    async def _process_chain_stream_update(
            self, chain_stream_update: Dict[str, Any], derivative_markets: List[InjectiveDerivativeMarket],
    ):
        self._last_received_message_timestamp = self._time()
        await super()._process_chain_stream_update(
            chain_stream_update=chain_stream_update,
            derivative_markets=derivative_markets,
        )

    async def _process_transaction_update(self, transaction_event: Dict[str, Any]):
        self._last_received_message_timestamp = self._time()
        await super()._process_transaction_update(transaction_event=transaction_event)

    async def _configure_gas_fee_for_transaction(self, transaction: Transaction):
        multiplier = (None
                      if CONSTANTS.GAS_LIMIT_ADJUSTMENT_MULTIPLIER is None
                      else Decimal(str(CONSTANTS.GAS_LIMIT_ADJUSTMENT_MULTIPLIER)))
        if self._fee_calculator is None:
            self._fee_calculator = self._fee_calculator_mode.create_calculator(
                client=self._client,
                composer=await self.composer(),
                gas_price=CONSTANTS.TX_GAS_PRICE,
                gas_limit_adjustment_multiplier=multiplier,
            )

        await self._fee_calculator.configure_gas_fee_for_transaction(
            transaction=transaction,
            private_key=self._private_key,
            public_key=self._public_key,
        )



import logging
import uuid
from copy import deepcopy
from decimal import Decimal
from typing import Dict, List

from hummingbot.connector.markets_recorder import MarketsRecorder
from hummingbot.core.data_type.common import PositionAction, PositionMode, PriceType, TradeType
from hummingbot.logger import HummingbotLogger
from hummingbot.model.position import Position
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.executors.arbitrage_executor.arbitrage_executor import ArbitrageExecutor
from hummingbot.strategy_v2.executors.data_types import PositionSummary
from hummingbot.strategy_v2.executors.dca_executor.dca_executor import DCAExecutor
from hummingbot.strategy_v2.executors.grid_executor.grid_executor import GridExecutor
from hummingbot.strategy_v2.executors.order_executor.order_executor import OrderExecutor
from hummingbot.strategy_v2.executors.position_executor.position_executor import PositionExecutor
from hummingbot.strategy_v2.executors.twap_executor.twap_executor import TWAPExecutor
from hummingbot.strategy_v2.executors.xemm_executor.xemm_executor import XEMMExecutor
from hummingbot.strategy_v2.models.executor_actions import (
    CreateExecutorAction,
    ExecutorAction,
    StopExecutorAction,
    StoreExecutorAction,
)
from hummingbot.strategy_v2.models.executors import CloseType
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo, PerformanceReport


class PositionHold:
    def __init__(self, connector_name: str, trading_pair: str, side: TradeType):
        self.connector_name = connector_name
        self.trading_pair = trading_pair
        self.side = side
        # Store only client order IDs
        self.order_ids = set()

        # Pre-calculated metrics
        self.volume_traded_quote = Decimal("0")
        self.cum_fees_quote = Decimal("0")

        # Separate tracking for buys and sells
        self.buy_amount_base = Decimal("0")
        self.buy_amount_quote = Decimal("0")
        self.sell_amount_base = Decimal("0")
        self.sell_amount_quote = Decimal("0")

    def add_orders_from_executor(self, executor: ExecutorInfo):
        custom_info = executor.custom_info
        if "held_position_orders" not in custom_info or len(custom_info["held_position_orders"]) == 0:
            return

        for order in custom_info["held_position_orders"]:
            # Skip if we've already processed this order
            order_id = order.get("client_order_id")
            if order_id in self.order_ids:
                continue

            # Add the order ID to our set
            self.order_ids.add(order_id)

            # Update metrics incrementally
            executed_amount_base = Decimal(str(order.get("executed_amount_base", 0)))
            executed_amount_quote = Decimal(str(order.get("executed_amount_quote", 0)))
            is_buy = order.get("trade_type") == "BUY"

            # Update volume traded in quote
            self.volume_traded_quote += executed_amount_quote

            # Update buy/sell amounts
            if is_buy:
                self.buy_amount_base += executed_amount_base
                self.buy_amount_quote += executed_amount_quote
            else:
                self.sell_amount_base += executed_amount_base
                self.sell_amount_quote += executed_amount_quote

            # Update fees
            self.cum_fees_quote += Decimal(str(order.get("cumulative_fee_paid_quote", 0)))

    def get_position_summary(self, mid_price: Decimal):
        # Calculate buy and sell breakeven prices
        buy_breakeven_price = self.buy_amount_quote / self.buy_amount_base if self.buy_amount_base > 0 else Decimal("0")
        sell_breakeven_price = self.sell_amount_quote / self.sell_amount_base if self.sell_amount_base > 0 else Decimal(
            "0")

        # Calculate matched volume (minimum of buy and sell base amounts)
        matched_amount_base = min(self.buy_amount_base, self.sell_amount_base)

        # Calculate realized PnL from matched volume
        realized_pnl_quote = (sell_breakeven_price - buy_breakeven_price) * matched_amount_base if matched_amount_base > 0 else Decimal("0")

        # Calculate net position amount and direction
        net_amount_base = self.buy_amount_base - self.sell_amount_base
        is_net_long = net_amount_base >= 0

        # Calculate unrealized PnL for the remaining unmatched volume
        unrealized_pnl_quote = Decimal("0")
        breakeven_price = Decimal("0")
        if net_amount_base != 0:
            if is_net_long:
                # Long position: remaining buy amount
                remaining_base = net_amount_base
                remaining_quote = self.buy_amount_quote - (matched_amount_base * buy_breakeven_price)
                breakeven_price = remaining_quote / remaining_base
                unrealized_pnl_quote = (mid_price - breakeven_price) * remaining_base
            else:
                # Short position: remaining sell amount
                remaining_base = abs(net_amount_base)
                remaining_quote = self.sell_amount_quote - (matched_amount_base * sell_breakeven_price)
                breakeven_price = remaining_quote / remaining_base
                unrealized_pnl_quote = (breakeven_price - mid_price) * remaining_base

        return PositionSummary(
            connector_name=self.connector_name,
            trading_pair=self.trading_pair,
            volume_traded_quote=self.volume_traded_quote,
            amount=abs(net_amount_base),
            side=TradeType.BUY if is_net_long else TradeType.SELL,
            breakeven_price=breakeven_price,
            unrealized_pnl_quote=unrealized_pnl_quote,
            realized_pnl_quote=realized_pnl_quote,
            cum_fees_quote=self.cum_fees_quote)


class ExecutorOrchestrator:
    """
    Orchestrator for various executors.
    """
    _logger = None
    _executor_mapping = {
        "position_executor": PositionExecutor,
        "grid_executor": GridExecutor,
        "dca_executor": DCAExecutor,
        "arbitrage_executor": ArbitrageExecutor,
        "twap_executor": TWAPExecutor,
        "xemm_executor": XEMMExecutor,
        "order_executor": OrderExecutor,
    }

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self,
                 strategy: ScriptStrategyBase,
                 executors_update_interval: float = 1.0,
                 executors_max_retries: int = 10):
        self.strategy = strategy
        self.executors_update_interval = executors_update_interval
        self.executors_max_retries = executors_max_retries
        self.active_executors = {}
        self.archived_executors = {}
        self.positions_held = {}
        self.executors_ids_position_held = []
        self.cached_performance = {}
        self._initialize_cached_performance()

    def _initialize_cached_performance(self):
        """
        Initialize cached performance by querying the database for stored executors.
        """
        db_executors = MarketsRecorder.get_instance().get_all_executors()
        for executor in db_executors:
            controller_id = executor.controller_id
            if controller_id not in self.cached_performance:
                self.cached_performance[controller_id] = PerformanceReport()
                self.active_executors[controller_id] = []
                self.archived_executors[controller_id] = []
                self.positions_held[controller_id] = []
            self._update_cached_performance(controller_id, executor)

    def _update_cached_performance(self, controller_id: str, executor_info: ExecutorInfo):
        """
        Update the cached performance for a specific controller with an executor's information.
        """
        report = self.cached_performance[controller_id]
        report.realized_pnl_quote += executor_info.net_pnl_quote
        report.volume_traded += executor_info.filled_amount_quote
        if executor_info.close_type:
            report.close_type_counts[executor_info.close_type] = report.close_type_counts.get(executor_info.close_type,
                                                                                              0) + 1

    def stop(self):
        """
        Stop the orchestrator task and all active executors.
        """
        # first we stop all active executors
        for controller_id, executors_list in self.active_executors.items():
            for executor in executors_list:
                if not executor.is_closed:
                    executor.early_stop()
        # Store all positions
        self.store_all_positions()

    def store_all_positions(self):
        """
        Store all positions in the database.
        """
        markets_recorder = MarketsRecorder.get_instance()
        for controller_id, positions_list in self.positions_held.items():
            for position in positions_list:
                mid_price = self.strategy.market_data_provider.get_price_by_type(
                    position.connector_name, position.trading_pair, PriceType.MidPrice)
                position_summary = position.get_position_summary(mid_price)

                # Create a new Position record
                position_record = Position(
                    id=str(uuid.uuid4()),
                    controller_id=controller_id,
                    connector_name=position_summary.connector_name,
                    trading_pair=position_summary.trading_pair,
                    side=position_summary.side.name,
                    timestamp=int(self.strategy.current_timestamp * 1e3),
                    volume_traded_quote=position_summary.volume_traded_quote,
                    amount=position_summary.amount,
                    breakeven_price=position_summary.breakeven_price,
                    unrealized_pnl_quote=position_summary.unrealized_pnl_quote,
                    cum_fees_quote=position_summary.cum_fees_quote,
                )
                # Store the position in the database
                markets_recorder.store_position(position_record)
                # Remove the position from the list
                self.positions_held[controller_id].remove(position)

    def store_all_executors(self):
        for controller_id, executors_list in self.active_executors.items():
            for executor in executors_list:
                # Store the executor in the database
                MarketsRecorder.get_instance().store_or_update_executor(executor)
        # Remove the executors from the list
        self.active_executors = {}

    def execute_action(self, action: ExecutorAction):
        """
        Execute the action and handle executors based on action type.
        """
        controller_id = action.controller_id
        if controller_id not in self.cached_performance:
            self.active_executors[controller_id] = []
            self.archived_executors[controller_id] = []
            self.positions_held[controller_id] = []
            self.cached_performance[controller_id] = PerformanceReport()

        if isinstance(action, CreateExecutorAction):
            self.create_executor(action)
        elif isinstance(action, StopExecutorAction):
            self.stop_executor(action)
        elif isinstance(action, StoreExecutorAction):
            self.store_executor(action)

    def execute_actions(self, actions: List[ExecutorAction]):
        """
        Execute a list of actions.
        """
        for action in actions:
            self.execute_action(action)

    def create_executor(self, action: CreateExecutorAction):
        """
        Create an executor based on the configuration in the action.
        """
        controller_id = action.controller_id
        executor_config = action.executor_config

        # For now, we replace the controller ID in the executor config with the actual controller object to mantain
        # compa
        executor_config.controller_id = controller_id

        executor_class = self._executor_mapping.get(executor_config.type)
        if executor_class is not None:
            executor = executor_class(
                strategy=self.strategy,
                config=executor_config,
                update_interval=self.executors_update_interval,
                max_retries=self.executors_max_retries,
            )
        else:
            raise ValueError("Unsupported executor config type")

        executor.start()
        self.active_executors[controller_id].append(executor)
        # MarketsRecorder.get_instance().store_or_update_executor(executor)
        self.logger().debug(f"Created {type(executor).__name__} for controller {controller_id}")

    def stop_executor(self, action: StopExecutorAction):
        """
        Stop an executor based on the action details.
        """
        controller_id = action.controller_id
        executor_id = action.executor_id

        executor = next(
            (executor for executor in self.active_executors[controller_id] if executor.config.id == executor_id),
            None)
        if not executor:
            self.logger().error(f"Executor ID {executor_id} not found for controller {controller_id}.")
            return
        executor.early_stop(action.keep_position)

    def store_executor(self, action: StoreExecutorAction):
        """
        Store executor data based on the action details and update cached performance.
        """
        controller_id = action.controller_id
        executor_id = action.executor_id

        executor = next(
            (executor for executor in self.active_executors[controller_id] if executor.config.id == executor_id),
            None)
        if not executor:
            self.logger().error(f"Executor ID {executor_id} not found for controller {controller_id}.")
            return
        if executor.is_active:
            self.logger().error(f"Executor ID {executor_id} is still active.")
            return
        try:
            MarketsRecorder.get_instance().store_or_update_executor(executor)
            self._update_cached_performance(controller_id, executor.executor_info)
        except Exception as e:
            self.logger().error(f"Error storing executor id {executor_id}: {str(e)}.")
            self.logger().error(f"Executor info: {executor.executor_info} | Config: {executor.config}")

        self.active_executors[controller_id].remove(executor)
        self.archived_executors[controller_id].append(executor.executor_info)
        del executor

    def get_executors_report(self) -> Dict[str, List[ExecutorInfo]]:
        """
        Generate a report of all executors.
        """
        report = {}
        for controller_id, executors_list in self.active_executors.items():
            report[controller_id] = [executor.executor_info for executor in executors_list if executor]
        return report

    def get_positions_report(self) -> Dict[str, List[PositionHold]]:
        """
        Generate a report of all positions held.
        """
        report = {}
        for controller_id, positions_list in self.positions_held.items():
            positions_summary = []
            for position in positions_list:
                mid_price = self.strategy.market_data_provider.get_price_by_type(
                    position.connector_name, position.trading_pair, PriceType.MidPrice)
                positions_summary.append(position.get_position_summary(mid_price))
            report[controller_id] = positions_summary
        return report

    def generate_performance_report(self, controller_id: str) -> PerformanceReport:
        # Start with a deep copy of the cached performance for this controller
        report = deepcopy(self.cached_performance.get(controller_id, PerformanceReport()))

        # Add data from active executors
        active_executors = self.active_executors.get(controller_id, [])
        positions = self.positions_held.get(controller_id, [])
        for executor in active_executors:
            executor_info = executor.executor_info
            side = executor_info.custom_info.get("side", None)
            if not executor_info.is_done:
                report.unrealized_pnl_quote += executor_info.net_pnl_quote
                if side:
                    report.inventory_imbalance += executor_info.filled_amount_quote \
                        if side == TradeType.BUY else -executor_info.filled_amount_quote
                if executor_info.type == "dca_executor":
                    report.open_order_volume += sum(
                        executor_info.config.amounts_quote) - executor_info.filled_amount_quote
                elif executor_info.type == "position_executor":
                    report.open_order_volume += (executor_info.config.amount *
                                                 executor_info.config.entry_price) - executor_info.filled_amount_quote
            else:
                report.realized_pnl_quote += executor_info.net_pnl_quote
                if executor_info.close_type in report.close_type_counts:
                    report.close_type_counts[executor_info.close_type] += 1
                else:
                    report.close_type_counts[executor_info.close_type] = 1
                if executor_info.close_type == CloseType.POSITION_HOLD and executor_info.config.id not in self.executors_ids_position_held:
                    self.executors_ids_position_held.append(executor_info.config.id)
                    # Check if this is a perpetual market
                    is_perpetual = "_perpetual" in executor_info.connector_name
                    # Get the position mode from the market
                    position_mode = None
                    position_side = None
                    if is_perpetual:
                        market = self.strategy.connectors[executor_info.connector_name]
                        if hasattr(market, 'position_mode'):
                            position_mode = market.position_mode
                        if hasattr(executor_info.config, "position_action") and position_mode == PositionMode.HEDGE:
                            opposite_side = TradeType.BUY if executor_info.config.side == TradeType.SELL else TradeType.SELL
                            position_side = opposite_side if executor_info.config.position_action == PositionAction.CLOSE else executor_info.config.side
                        else:
                            position_side = executor_info.config.side

                    if position_side:
                        # Find existing position for this trading pair
                        existing_position = next(
                            (position for position in positions if
                             position.trading_pair == executor_info.trading_pair and
                             position.connector_name == executor_info.connector_name and
                             position.side == position_side), None
                        )
                    else:
                        # Find existing position for this trading pair
                        existing_position = next(
                            (position for position in positions if
                             position.trading_pair == executor_info.trading_pair and
                             position.connector_name == executor_info.connector_name), None
                        )
                    if existing_position:
                        existing_position.add_orders_from_executor(executor_info)
                    else:
                        # Create new position
                        position = PositionHold(
                            executor_info.connector_name,
                            executor_info.trading_pair,
                            executor_info.config.side
                        )
                        position.add_orders_from_executor(executor_info)
                        positions.append(position)

            report.volume_traded += executor_info.filled_amount_quote

        # Add data from positions held
        for position in positions:
            mid_price = self.strategy.market_data_provider.get_price_by_type(
                position.connector_name, position.trading_pair, PriceType.MidPrice)
            position_summary = position.get_position_summary(mid_price)

            # Update report with position data
            report.realized_pnl_quote += position_summary.realized_pnl_quote - position_summary.cum_fees_quote
            report.volume_traded += position_summary.volume_traded_quote
            report.inventory_imbalance += position_summary.amount  # This is the net position amount
            report.unrealized_pnl_quote += position_summary.unrealized_pnl_quote
            # Store position summary in report for controller access
            if not hasattr(report, "positions_summary"):
                report.positions_summary = []
            report.positions_summary.append(position_summary)

        # Calculate global PNL values
        report.global_pnl_quote = report.unrealized_pnl_quote + report.realized_pnl_quote
        report.global_pnl_pct = (report.global_pnl_quote / report.volume_traded) * 100 if report.volume_traded != 0 else Decimal(0)

        # Calculate individual PNL percentages
        report.unrealized_pnl_pct = (report.unrealized_pnl_quote / report.volume_traded) * 100 if report.volume_traded != 0 else Decimal(0)
        report.realized_pnl_pct = (report.realized_pnl_quote / report.volume_traded) * 100 if report.volume_traded != 0 else Decimal(0)

        return report



#!/usr/bin/env python

import io
import os
import sys
import time
import traceback
from logging import Logger as PythonLogger
from typing import Optional, Type

import pandas as pd

from .application_warning import ApplicationWarning

TESTING_TOOLS = ["unittest", "pytest"]

#  --- Copied from logging module ---
if hasattr(sys, '_getframe'):
    def currentframe():
        return sys._getframe(3)
else:   # pragma: no cover
    def currentframe():
        """Return the frame object for the caller's stack frame."""
        try:
            raise Exception
        except Exception:
            return sys.exc_info()[2].tb_frame.f_back
#  --- Copied from logging module ---


class HummingbotLogger(PythonLogger):
    def __init__(self, name: str):
        super().__init__(name)

    @staticmethod
    def logger_name_for_class(model_class: Type):
        return f"{model_class.__module__}.{model_class.__qualname__}"

    @staticmethod
    def is_testing_mode() -> bool:
        return any(tools in arg
                   for tools in TESTING_TOOLS
                   for arg in sys.argv)

    def notify(self, msg: str):
        from . import INFO
        self.log(INFO, msg)
        if not HummingbotLogger.is_testing_mode():
            from hummingbot.client.hummingbot_application import HummingbotApplication
            hummingbot_app: HummingbotApplication = HummingbotApplication.main_application()
            hummingbot_app.notify(f"({pd.Timestamp.fromtimestamp(int(time.time()))}) {msg}")

    def network(self, log_msg: str, app_warning_msg: Optional[str] = None, *args, **kwargs):
        if app_warning_msg is not None and not HummingbotLogger.is_testing_mode():
            from hummingbot.client.hummingbot_application import HummingbotApplication

        from . import NETWORK

        self.log(NETWORK, log_msg, *args, **kwargs)
        if app_warning_msg is not None and not HummingbotLogger.is_testing_mode():
            app_warning: ApplicationWarning = ApplicationWarning(
                time.time(),
                self.name,
                self.findCaller(),
                app_warning_msg
            )
            self.warning(app_warning.warning_msg)
            hummingbot_app: HummingbotApplication = HummingbotApplication.main_application()
            hummingbot_app.add_application_warning(app_warning)

    #  --- Copied from logging module ---
    def findCaller(self, stack_info=False, stacklevel=1):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        orig_f = f
        while f and stacklevel > 1:
            f = f.f_back
            stacklevel -= 1
        if not f:
            f = orig_f
        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _srcfile:
                f = f.f_back
                continue
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write('Stack (most recent call last):\n')
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == '\n':
                    sinfo = sinfo[:-1]
                sio.close()
            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break
        return rv
    #  --- Copied from logging module ---


_srcfile = os.path.normcase(HummingbotLogger.network.__code__.co_filename)



import math
from datetime import datetime
from decimal import Decimal
from typing import Optional

from hummingbot.client.config.config_validators import (
    validate_bool,
    validate_datetime_iso_string,
    validate_decimal,
    validate_exchange,
    validate_market_trading_pair,
)
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.settings import AllConnectorSettings, required_exchanges


def trading_pair_prompt():
    exchange = twap_config_map.get("connector").value
    example = AllConnectorSettings.get_example_pairs().get(exchange)
    return "Enter the token trading pair you would like to trade on %s%s >>> " \
           % (exchange, f" (e.g. {example})" if example else "")


def target_asset_amount_prompt():
    trading_pair = twap_config_map.get("trading_pair").value
    base_token, _ = trading_pair.split("-")

    return f"What is the total amount of {base_token} to be traded? (Default is 1.0) >>> "


def str2bool(value: str):
    return str(value).lower() in ("yes", "y", "true", "t", "1")


# checks if the trading pair is valid
def validate_market_trading_pair_tuple(value: str) -> Optional[str]:
    exchange = twap_config_map.get("connector").value
    return validate_market_trading_pair(exchange, value)


def set_order_delay_default(value: str = None):
    start_datetime_string = twap_config_map.get("start_datetime").value
    end_datetime_string = twap_config_map.get("end_datetime").value
    start_datetime = datetime.fromisoformat(start_datetime_string)
    end_datetime = datetime.fromisoformat(end_datetime_string)

    target_asset_amount = twap_config_map.get("target_asset_amount").value
    order_step_size = twap_config_map.get("order_step_size").value

    default = math.floor((end_datetime - start_datetime).total_seconds() / math.ceil(target_asset_amount / order_step_size))
    twap_config_map.get("order_delay_time").default = default


def validate_order_step_size(value: str = None):
    """
    Invalidates non-decimal input and checks if order_step_size is less than the target_asset_amount value
    :param value: User input for order_step_size parameter
    :return: Error message printed in output pane
    """
    result = validate_decimal(value, min_value=Decimal("0"), inclusive=False)
    if result is not None:
        return result
    target_asset_amount = twap_config_map.get("target_asset_amount").value
    if Decimal(value) > target_asset_amount:
        return "Order step size cannot be greater than the total trade amount."


twap_config_map = {
    "strategy":
        ConfigVar(key="strategy",
                  prompt=None,
                  default="twap"),
    "connector":
        ConfigVar(key="connector",
                  prompt="Enter the name of spot connector >>> ",
                  validator=validate_exchange,
                  on_validated=lambda value: required_exchanges.add(value),
                  prompt_on_new=True),
    "trading_pair":
        ConfigVar(key="trading_pair",
                  prompt=trading_pair_prompt,
                  validator=validate_market_trading_pair_tuple,
                  prompt_on_new=True),
    "trade_side":
        ConfigVar(key="trade_side",
                  prompt="What operation will be executed? (buy/sell) >>> ",
                  type_str="str",
                  validator=lambda v: None if v in {"buy", "sell", ""} else "Invalid operation type.",
                  default="buy",
                  prompt_on_new=True),
    "target_asset_amount":
        ConfigVar(key="target_asset_amount",
                  prompt=target_asset_amount_prompt,
                  default=1.0,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=Decimal("0"), inclusive=False),
                  prompt_on_new=True),
    "order_step_size":
        ConfigVar(key="order_step_size",
                  prompt="What is the amount of each individual order (denominated in the base asset, default is 1)? "
                         ">>> ",
                  default=1.0,
                  type_str="decimal",
                  validator=validate_order_step_size,
                  prompt_on_new=True),
    "order_price":
        ConfigVar(key="order_price",
                  prompt="What is the price for the limit orders? >>> ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=Decimal("0"), inclusive=False),
                  prompt_on_new=True),
    "is_delayed_start_execution":
        ConfigVar(key="is_delayed_start_execution",
                  prompt="Do you want to specify a start time for the execution? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool,
                  prompt_on_new=True),
    "start_datetime":
        ConfigVar(key="start_datetime",
                  prompt="Please enter the start date and time"
                         " (YYYY-MM-DD HH:MM:SS) >>> ",
                  type_str="str",
                  validator=validate_datetime_iso_string,
                  required_if=lambda: twap_config_map.get("is_time_span_execution").value or twap_config_map.get("is_delayed_start_execution").value,
                  prompt_on_new=True),
    "is_time_span_execution":
        ConfigVar(key="is_time_span_execution",
                  prompt="Do you want to specify an end time for the execution? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool,
                  prompt_on_new=True),
    "end_datetime":
        ConfigVar(key="end_datetime",
                  prompt="Please enter the end date and time"
                         " (YYYY-MM-DD HH:MM:SS) >>> ",
                  type_str="str",
                  validator=validate_datetime_iso_string,
                  on_validated=set_order_delay_default,
                  required_if=lambda: twap_config_map.get("is_time_span_execution").value,
                  prompt_on_new=True),
    "order_delay_time":
        ConfigVar(key="order_delay_time",
                  prompt="How many seconds do you want to wait between each individual order?"
                         " (Enter 10 to indicate 10 seconds)? >>> ",
                  type_str="float",
                  default=10,
                  validator=lambda v: validate_decimal(v, 0, inclusive=False),
                  required_if=lambda: twap_config_map.get("is_time_span_execution").value or twap_config_map.get("is_delayed_start_execution").value,
                  prompt_on_new=True),
    "cancel_order_wait_time":
        ConfigVar(key="cancel_order_wait_time",
                  prompt="How long do you want to wait before canceling your limit order (in seconds). "
                         "(Default is 60 seconds) ? >>> ",
                  type_str="float",
                  default=60,
                  validator=lambda v: validate_decimal(v, 0, inclusive=False),
                  prompt_on_new=True)
}



import datetime
import os
from decimal import Decimal
from typing import Dict, List, Set

import pandas as pd
import pandas_ta as ta  # noqa: F401

from hummingbot import data_path
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.executors.position_executor.data_types import (
    PositionExecutorConfig,
    TrailingStop,
    TripleBarrierConfig,
)
from hummingbot.strategy_v2.executors.position_executor.position_executor import PositionExecutor


class DirectionalStrategyBase(ScriptStrategyBase):
    """
    Base class to create directional strategies using the PositionExecutor.

    Attributes:
        directional_strategy_name (str): The name of the directional strategy.
        trading_pair (str): The trading pair to be used.
        exchange (str): The exchange to be used.
        max_executors (int): Maximum number of position executors to be active at a time.
        position_mode (PositionMode): The position mode to be used.
        active_executors (List[PositionExecutor]): List of currently active position executors.
        stored_executors (List[PositionExecutor]): List of closed position executors that have been stored.
        stop_loss (float): The stop loss percentage.
        take_profit (float): The take profit percentage.
        time_limit (int): The time limit for the position in seconds.
        open_order_type (OrderType): The order type for opening the position.
        open_order_slippage_buffer (float): The slippage buffer for the opening order.
        take_profit_order_type (OrderType): The order type for the take profit order.
        stop_loss_order_type (OrderType): The order type for the stop loss order.
        time_limit_order_type (OrderType): The order type for the time limit order.
        trailing_stop_activation_delta (float): The delta for activating the trailing stop.
        trailing_stop_trailing_delta (float): The delta for trailing the stop loss.
        candles (List[CandlesBase]): List of candlestick data sources to be used.
        set_leverage_flag (None): Flag indicating whether leverage has been set.
        leverage (float): The leverage to be used.
        order_amount_usd (Decimal): The order amount in USD.
        markets (Dict[str, Set[str]]): Dictionary mapping exchanges to trading pairs.
        cooldown_after_execution (int): Cooldown between position executions, in seconds.
    """
    directional_strategy_name: str
    # Define the trading pair and exchange that we want to use and the csv where we are going to store the entries
    trading_pair: str
    exchange: str

    # Maximum position executors at a time
    max_executors: int = 1
    position_mode: PositionMode = PositionMode.HEDGE
    active_executors: List[PositionExecutor] = []
    stored_executors: List[PositionExecutor] = []

    # Configure the parameters for the position
    stop_loss: float = 0.01
    take_profit: float = 0.01
    time_limit: int = 120
    open_order_type = OrderType.MARKET
    open_order_slippage_buffer: float = 0.001
    take_profit_order_type: OrderType = OrderType.MARKET
    stop_loss_order_type: OrderType = OrderType.MARKET
    time_limit_order_type: OrderType = OrderType.MARKET
    trailing_stop_activation_delta = 0.003
    trailing_stop_trailing_delta = 0.001
    cooldown_after_execution = 30

    # Create the candles that we want to use and the thresholds for the indicators
    candles: List[CandlesBase]

    # Configure the leverage and order amount the bot is going to use
    set_leverage_flag = None
    leverage = 1
    order_amount_usd = Decimal("10")
    markets: Dict[str, Set[str]] = {}

    @property
    def all_candles_ready(self):
        """
        Checks if the candlesticks are full.
        """
        return all([candle.ready for candle in self.candles])

    @property
    def is_perpetual(self):
        """
        Checks if the exchange is a perpetual market.
        """
        return "perpetual" in self.exchange

    @property
    def max_active_executors_condition(self):
        return len(self.get_active_executors()) < self.max_executors

    @property
    def time_between_signals_condition(self):
        seconds_since_last_signal = self.current_timestamp - self.get_timestamp_of_last_executor()
        return seconds_since_last_signal > self.cooldown_after_execution

    def get_csv_path(self) -> str:
        today = datetime.datetime.today()
        csv_path = data_path() + f"/{self.directional_strategy_name}_position_executors_{self.exchange}_{self.trading_pair}_{today.day:02d}-{today.month:02d}-{today.year}.csv"
        return csv_path

    def __init__(self, connectors: Dict[str, ConnectorBase]):
        # Is necessary to start the Candles Feed.
        super().__init__(connectors)
        self.triple_barrier_conf = TripleBarrierConfig(
            stop_loss=Decimal(self.stop_loss),
            take_profit=Decimal(self.take_profit),
            time_limit=self.time_limit,
            trailing_stop=TrailingStop(
                activation_price=Decimal(self.trailing_stop_activation_delta),
                trailing_delta=Decimal(self.trailing_stop_trailing_delta)),
            open_order_type=self.open_order_type,
            take_profit_order_type=self.take_profit_order_type,
            stop_loss_order_type=self.stop_loss_order_type,
            time_limit_order_type=self.time_limit_order_type
        )
        for candle in self.candles:
            candle.start()

    def candles_formatted_list(self, candles_df: pd.DataFrame, columns_to_show: List):
        lines = []
        candles_df = candles_df.copy()
        candles_df["timestamp"] = pd.to_datetime(candles_df["timestamp"], unit="ms")
        lines.extend(["    " + line for line in candles_df[columns_to_show].tail().to_string(index=False).split("\n")])
        lines.extend(["\n-----------------------------------------------------------------------------------------------------------\n"])
        return lines

    def on_stop(self):
        """
        Without this functionality, the network iterator will continue running forever after stopping the strategy
        That's why is necessary to introduce this new feature to make a custom stop with the strategy.
        """
        if self.is_perpetual:
            # we are going to close all the open positions when the bot stops
            self.close_open_positions()
        for candle in self.candles:
            candle.stop()

    def get_active_executors(self) -> List[PositionExecutor]:
        return [signal_executor for signal_executor in self.active_executors
                if not signal_executor.is_closed]

    def get_closed_executors(self) -> List[PositionExecutor]:
        return [signal_executor for signal_executor in self.active_executors
                if signal_executor.is_closed]

    def get_timestamp_of_last_executor(self):
        if len(self.stored_executors) > 0:
            return self.stored_executors[-1].close_timestamp
        else:
            return 0

    def on_tick(self):
        self.clean_and_store_executors()
        if self.is_perpetual:
            self.check_and_set_leverage()
        if self.max_active_executors_condition and self.all_candles_ready and self.time_between_signals_condition:
            position_config = self.get_position_config()
            if position_config:
                signal_executor = PositionExecutor(
                    strategy=self,
                    config=position_config,
                )
                self.active_executors.append(signal_executor)

    def get_position_config(self):
        signal = self.get_signal()
        if signal == 0:
            return None
        else:
            price = self.connectors[self.exchange].get_mid_price(self.trading_pair)
            side = TradeType.BUY if signal == 1 else TradeType.SELL
            if self.open_order_type.is_limit_type():
                price = price * (1 - signal * self.open_order_slippage_buffer)
            position_config = PositionExecutorConfig(
                timestamp=self.current_timestamp,
                trading_pair=self.trading_pair,
                connector_name=self.exchange,
                side=side,
                amount=self.order_amount_usd / price,
                entry_price=price,
                triple_barrier_config=self.triple_barrier_conf,
                leverage=self.leverage,
            )
            return position_config

    def get_signal(self):
        """Base method to get the signal from the candles."""
        raise NotImplementedError

    def format_status(self) -> str:
        """
        Displays the three candlesticks involved in the script with RSI, BBANDS and EMA.
        """
        if not self.ready_to_trade:
            return "Market connectors are not ready."
        lines = []

        if len(self.stored_executors) > 0:
            lines.extend(["\n################################## Closed Executors ##################################"])
        for executor in self.stored_executors:
            lines.extend([f"|Signal id: {executor.config.timestamp}"])
            lines.extend(executor.to_format_status())
            lines.extend([
                "-----------------------------------------------------------------------------------------------------------"])

        if len(self.active_executors) > 0:
            lines.extend(["\n################################## Active Executors ##################################"])

        for executor in self.active_executors:
            lines.extend([f"|Signal id: {executor.config.timestamp}"])
            lines.extend(executor.to_format_status())
        if self.all_candles_ready:
            lines.extend(["\n################################## Market Data ##################################\n"])
            lines.extend([f"Value: {self.get_signal()}"])
            lines.extend(self.market_data_extra_info())
        else:
            lines.extend(["", "  No data collected."])

        return "\n".join(lines)

    def check_and_set_leverage(self):
        if not self.set_leverage_flag:
            for connector in self.connectors.values():
                for trading_pair in connector.trading_pairs:
                    connector.set_position_mode(self.position_mode)
                    connector.set_leverage(trading_pair=trading_pair, leverage=self.leverage)
            self.set_leverage_flag = True

    def clean_and_store_executors(self):
        executors_to_store = [executor for executor in self.active_executors if executor.is_closed]
        csv_path = self.get_csv_path()
        if not os.path.exists(csv_path):
            df_header = pd.DataFrame([("timestamp",
                                       "exchange",
                                       "trading_pair",
                                       "side",
                                       "amount",
                                       "trade_pnl",
                                       "trade_pnl_quote",
                                       "cum_fee_quote",
                                       "net_pnl_quote",
                                       "net_pnl",
                                       "close_timestamp",
                                       "close_type",
                                       "entry_price",
                                       "close_price",
                                       "sl",
                                       "tp",
                                       "tl",
                                       "open_order_type",
                                       "take_profit_order_type",
                                       "stop_loss_order_type",
                                       "time_limit_order_type",
                                       "leverage"
                                       )])
            df_header.to_csv(csv_path, mode='a', header=False, index=False)
        for executor in executors_to_store:
            self.stored_executors.append(executor)
            df = pd.DataFrame([(executor.config.timestamp,
                                executor.config.connector_name,
                                executor.config.trading_pair,
                                executor.config.side,
                                executor.config.amount,
                                executor.trade_pnl_pct,
                                executor.trade_pnl_quote,
                                executor.cum_fees_quote,
                                executor.net_pnl_quote,
                                executor.net_pnl_pct,
                                executor.close_timestamp,
                                executor.close_type,
                                executor.entry_price,
                                executor.close_price,
                                executor.config.triple_barrier_config.stop_loss,
                                executor.config.triple_barrier_config.take_profit,
                                executor.config.triple_barrier_config.time_limit,
                                executor.config.triple_barrier_config.open_order_type,
                                executor.config.triple_barrier_config.take_profit_order_type,
                                executor.config.triple_barrier_config.stop_loss_order_type,
                                executor.config.triple_barrier_config.time_limit_order_type,
                                self.leverage)])
            df.to_csv(self.get_csv_path(), mode='a', header=False, index=False)
        self.active_executors = [executor for executor in self.active_executors if not executor.is_closed]

    def close_open_positions(self):
        # we are going to close all the open positions when the bot stops
        for connector_name, connector in self.connectors.items():
            for trading_pair, position in connector.account_positions.items():
                if position.position_side == PositionSide.LONG:
                    self.sell(connector_name=connector_name,
                              trading_pair=position.trading_pair,
                              amount=abs(position.amount),
                              order_type=OrderType.MARKET,
                              price=connector.get_mid_price(position.trading_pair),
                              position_action=PositionAction.CLOSE)
                elif position.position_side == PositionSide.SHORT:
                    self.buy(connector_name=connector_name,
                             trading_pair=position.trading_pair,
                             amount=abs(position.amount),
                             order_type=OrderType.MARKET,
                             price=connector.get_mid_price(position.trading_pair),
                             position_action=PositionAction.CLOSE)

    def market_data_extra_info(self):
        return ["\n"]



import hashlib
import hmac
import json
from collections import OrderedDict
from typing import Any, Dict
from urllib.parse import urlencode

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class BinancePerpetualAuth(AuthBase):
    """
    Auth class required by Binance Perpetual API
    """

    def __init__(self, api_key: str, api_secret: str, time_provider: TimeSynchronizer):
        self._api_key: str = api_key
        self._api_secret: str = api_secret
        self._time_provider: TimeSynchronizer = time_provider

    def generate_signature_from_payload(self, payload: str) -> str:
        secret = bytes(self._api_secret.encode("utf-8"))
        signature = hmac.new(secret, payload.encode("utf-8"), hashlib.sha256).hexdigest()
        return signature

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        if request.method == RESTMethod.POST:
            request.data = self.add_auth_to_params(params=json.loads(request.data))
        else:
            request.params = self.add_auth_to_params(request.params)

        request.headers = self.header_for_authentication()

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request  # pass-through

    def add_auth_to_params(self,
                           params: Dict[str, Any]):
        timestamp = int(self._time_provider.time() * 1e3)

        request_params = OrderedDict(params or {})
        request_params["timestamp"] = timestamp

        payload = urlencode(request_params)
        request_params["signature"] = self.generate_signature_from_payload(payload=payload)

        return request_params

    def header_for_authentication(self) -> Dict[str, str]:
        return {"X-MBX-APIKEY": self._api_key}



from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# Kucoin Futures fees: https://www.kucoin.com/vip/level
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0002"),
    taker_percent_fee_decimal=Decimal("0.0006"),
    percent_fee_token="USDT")

CENTRALIZED = True

EXAMPLE_PAIR = "XBT-USDT"


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    status = exchange_info.get("status")
    valid = status is not None and status in ["Open"]
    return valid


class KucoinPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="kucoin_perpetual", client_data=None)
    kucoin_perpetual_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Kucoin Perpetual API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    kucoin_perpetual_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your Kucoin Perpetual secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    kucoin_perpetual_passphrase: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: "Enter your KuCoin Perpetual passphrase",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="kucoin_perpetual")


KEYS = KucoinPerpetualConfigMap.construct()
